import numpy as np
import torch
import torch.nn.functional as F
from torch.autograd import Variable
from torch.distributions import Distribution


class DiscretizedMixtureLogitsDistribution(Distribution):
    def __init__(self, nr_mix, logits):
        self.logits = logits
        self.nr_mix = nr_mix

    def log_prob(self, value):
        return - discretized_mix_logistic_loss(value, self.logits)

    def sample(self):
        return sample_from_discretized_mix_logistic(self.logits, self.nr_mix)


class DiscretizedMixtureLogits():

    def __init__(self, nr_mix, **kwargs):
        self.nr_mix = nr_mix

    def __call__(self, logits):
        return DiscretizedMixtureLogitsDistribution(self.nr_mix, logits)


# copied from: https://github.com/pclucas14/pixel-cnn-pp/blob/master/utils.py


def concat_elu(x):
    """ like concatenated ReLU (http://arxiv.org/abs/1603.05201), but then with ELU """
    # Pytorch ordering
    axis = len(x.size()) - 3
    return F.elu(torch.cat([x, -x], dim=axis))


def log_sum_exp(x):
    """ numerically stable log_sum_exp implementation that prevents overflow """
    # TF ordering
    axis = len(x.size()) - 1
    m, _ = torch.max(x, dim=axis)
    m2, _ = torch.max(x, dim=axis, keepdim=True)
    return m + torch.log(torch.sum(torch.exp(x - m2), dim=axis))


def log_prob_from_logits(x):
    """ numerically stable log_softmax implementation that prevents overflow """
    # TF ordering
    axis = len(x.size()) - 1
    m, _ = torch.max(x, dim=axis, keepdim=True)
    return x - m - torch.log(torch.sum(torch.exp(x - m), dim=axis, keepdim=True))


def discretized_mix_logistic_loss(x, l):
    """ log-likelihood for mixture of discretized logistics, assumes the data has been rescaled to [-1,1] interval """
    # Pytorch ordering
    x = x.permute(0, 2, 3, 1)
    l = l.permute(0, 2, 3, 1)
    xs = [int(y) for y in x.size()]
    ls = [int(y) for y in l.size()]

    # here and below: unpacking the params of the mixture of logistics
    nr_mix = int(ls[-1] / 10)
    logit_probs = l[:, :, :, :nr_mix]
    l = l[:, :, :, nr_mix:].contiguous().view(xs + [nr_mix * 3])  # 3 for mean, scale, coef
    means = l[:, :, :, :, :nr_mix]
    # log_scales = torch.max(l[:, :, :, :, nr_mix:2 * nr_mix], -7.)
    log_scales = torch.clamp(l[:, :, :, :, nr_mix:2 * nr_mix], min=-7.)

    coeffs = torch.tanh(l[:, :, :, :, 2 * nr_mix:3 * nr_mix])
    # here and below: getting the means and adjusting them based on preceding
    # sub-pixels
    x = x.contiguous()
    x = x.unsqueeze(-1) + torch.zeros(xs + [nr_mix], device=x.device)
    m2 = (means[:, :, :, 1, :] + coeffs[:, :, :, 0, :]
          * x[:, :, :, 0, :]).view(xs[0], xs[1], xs[2], 1, nr_mix)

    m3 = (means[:, :, :, 2, :] + coeffs[:, :, :, 1, :] * x[:, :, :, 0, :] +
          coeffs[:, :, :, 2, :] * x[:, :, :, 1, :]).view(xs[0], xs[1], xs[2], 1, nr_mix)

    means = torch.cat((means[:, :, :, 0, :].unsqueeze(3), m2, m3), dim=3)
    centered_x = x - means
    inv_stdv = torch.exp(-log_scales)
    plus_in = inv_stdv * (centered_x + 1. / 255.)
    cdf_plus = torch.sigmoid(plus_in)
    min_in = inv_stdv * (centered_x - 1. / 255.)
    cdf_min = torch.sigmoid(min_in)
    # log probability for edge case of 0 (before scaling)
    log_cdf_plus = plus_in - F.softplus(plus_in)
    # log probability for edge case of 255 (before scaling)
    log_one_minus_cdf_min = -F.softplus(min_in)
    cdf_delta = cdf_plus - cdf_min  # probability for all other cases
    mid_in = inv_stdv * centered_x
    # log probability in the center of the bin, to be used in extreme cases
    # (not actually used in our code)
    log_pdf_mid = mid_in - log_scales - 2. * F.softplus(mid_in)

    # now select the right output: left edge case, right edge case, normal
    # case, extremely low prob case (doesn't actually happen for us)

    # this is what we are really doing, but using the robust version below for extreme cases in other applications and to avoid NaN issue with tf.select()
    # log_probs = tf.select(x < -0.999, log_cdf_plus, tf.select(x > 0.999, log_one_minus_cdf_min, tf.log(cdf_delta)))

    # robust version, that still works if probabilities are below 1e-5 (which never happens in our code)
    # tensorflow backpropagates through tf.select() by multiplying with zero instead of selecting: this requires use to use some ugly tricks to avoid potential NaNs
    # the 1e-12 in tf.maximum(cdf_delta, 1e-12) is never actually used as output, it's purely there to get around the tf.select() gradient issue
    # if the probability on a sub-pixel is below 1e-5, we use an approximation
    # based on the assumption that the log-density is constant in the bin of
    # the observed sub-pixel value

    inner_inner_cond = (cdf_delta > 1e-5).float()
    inner_inner_out = inner_inner_cond * torch.log(torch.clamp(cdf_delta, min=1e-12)) + (1. - inner_inner_cond) * (
            log_pdf_mid - np.log(127.5))
    inner_cond = (x > 0.999).float()
    inner_out = inner_cond * log_one_minus_cdf_min + (1. - inner_cond) * inner_inner_out
    cond = (x < -0.999).float()
    log_probs = cond * log_cdf_plus + (1. - cond) * inner_out
    log_probs = torch.sum(log_probs, dim=3) + log_prob_from_logits(logit_probs)

    return - log_sum_exp(log_probs)


def discretized_mix_logistic_loss_1d(x, l):
    """ log-likelihood for mixture of discretized logistics, assumes the data has been rescaled to [-1,1] interval """
    # Pytorch ordering
    x = x.permute(0, 2, 3, 1)
    l = l.permute(0, 2, 3, 1)
    xs = [int(y) for y in x.size()]
    ls = [int(y) for y in l.size()]

    # here and below: unpacking the params of the mixture of logistics
    nr_mix = int(ls[-1] / 3)
    logit_probs = l[:, :, :, :nr_mix]
    l = l[:, :, :, nr_mix:].contiguous().view(xs + [nr_mix * 2])  # 2 for mean, scale
    means = l[:, :, :, :, :nr_mix]
    log_scales = torch.clamp(l[:, :, :, :, nr_mix:2 * nr_mix], min=-7.)
    # here and below: getting the means and adjusting them based on preceding
    # sub-pixels
    x = x.contiguous()
    x = x.unsqueeze(-1) + Variable(torch.zeros(xs + [nr_mix]).cuda(), requires_grad=False)

    # means = torch.cat((means[:, :, :, 0, :].unsqueeze(3), m2, m3), dim=3)
    centered_x = x - means
    inv_stdv = torch.exp(-log_scales)
    plus_in = inv_stdv * (centered_x + 1. / 255.)
    cdf_plus = torch.sigmoid(plus_in)
    min_in = inv_stdv * (centered_x - 1. / 255.)
    cdf_min = torch.sigmoid(min_in)
    # log probability for edge case of 0 (before scaling)
    log_cdf_plus = plus_in - F.softplus(plus_in)
    # log probability for edge case of 255 (before scaling)
    log_one_minus_cdf_min = -F.softplus(min_in)
    cdf_delta = cdf_plus - cdf_min  # probability for all other cases
    mid_in = inv_stdv * centered_x
    # log probability in the center of the bin, to be used in extreme cases
    # (not actually used in our code)
    log_pdf_mid = mid_in - log_scales - 2. * F.softplus(mid_in)

    inner_inner_cond = (cdf_delta > 1e-5).float()
    inner_inner_out = inner_inner_cond * torch.log(torch.clamp(cdf_delta, min=1e-12)) + (1. - inner_inner_cond) * (
            log_pdf_mid - np.log(127.5))
    inner_cond = (x > 0.999).float()
    inner_out = inner_cond * log_one_minus_cdf_min + (1. - inner_cond) * inner_inner_out
    cond = (x < -0.999).float()
    log_probs = cond * log_cdf_plus + (1. - cond) * inner_out
    log_probs = torch.sum(log_probs, dim=3) + log_prob_from_logits(logit_probs)

    return -torch.sum(log_sum_exp(log_probs))


def to_one_hot(tensor, n, fill_with=1.):
    # we perform one hot encore with respect to the last axis
    one_hot = torch.FloatTensor(tensor.size() + (n,)).zero_()
    if tensor.is_cuda: one_hot = one_hot.cuda()
    one_hot.scatter_(len(tensor.size()), tensor.unsqueeze(-1), fill_with)
    return Variable(one_hot)


def sample_from_discretized_mix_logistic_1d(l, nr_mix):
    # Pytorch ordering
    l = l.permute(0, 2, 3, 1)
    ls = [int(y) for y in l.size()]
    xs = ls[:-1] + [1]  # [3]

    # unpack parameters
    logit_probs = l[:, :, :, :nr_mix]
    l = l[:, :, :, nr_mix:].contiguous().view(xs + [nr_mix * 2])  # for mean, scale

    # sample mixture indicator from softmax
    temp = torch.FloatTensor(logit_probs.size())
    if l.is_cuda: temp = temp.cuda()
    temp.uniform_(1e-5, 1. - 1e-5)
    temp = logit_probs.data - torch.log(- torch.log(temp))
    _, argmax = temp.max(dim=3)

    one_hot = to_one_hot(argmax, nr_mix)
    sel = one_hot.view(xs[:-1] + [1, nr_mix])
    # select logistic parameters
    means = torch.sum(l[:, :, :, :, :nr_mix] * sel, dim=4)
    log_scales = torch.clamp(torch.sum(
        l[:, :, :, :, nr_mix:2 * nr_mix] * sel, dim=4), min=-7.)
    u = torch.FloatTensor(means.size())
    if l.is_cuda: u = u.cuda()
    u.uniform_(1e-5, 1. - 1e-5)
    u = Variable(u)
    x = means + torch.exp(log_scales) * (torch.log(u) - torch.log(1. - u))
    x0 = torch.clamp(torch.clamp(x[:, :, :, 0], min=-1.), max=1.)
    out = x0.unsqueeze(1)
    return out


def sample_from_discretized_mix_logistic(l, nr_mix):
    # Pytorch ordering
    l = l.permute(0, 2, 3, 1)
    ls = [int(y) for y in l.size()]
    xs = ls[:-1] + [3]

    # unpack parameters
    logit_probs = l[:, :, :, :nr_mix]
    l = l[:, :, :, nr_mix:].contiguous().view(xs + [nr_mix * 3])
    # sample mixture indicator from softmax
    temp = torch.FloatTensor(logit_probs.size())
    if l.is_cuda: temp = temp.cuda()
    temp.uniform_(1e-5, 1. - 1e-5)
    temp = logit_probs.data - torch.log(- torch.log(temp))
    _, argmax = temp.max(dim=3)

    one_hot = to_one_hot(argmax, nr_mix)
    sel = one_hot.view(xs[:-1] + [1, nr_mix])
    # select logistic parameters
    means = torch.sum(l[:, :, :, :, :nr_mix] * sel, dim=4)
    log_scales = torch.clamp(torch.sum(
        l[:, :, :, :, nr_mix:2 * nr_mix] * sel, dim=4), min=-7.)
    coeffs = torch.sum(torch.tanh(
        l[:, :, :, :, 2 * nr_mix:3 * nr_mix]) * sel, dim=4)
    # sample from logistic & clip to interval
    # we don't actually round to the nearest 8bit value when sampling
    u = torch.FloatTensor(means.size())
    if l.is_cuda: u = u.cuda()
    u.uniform_(1e-5, 1. - 1e-5)
    u = Variable(u)
    x = means + torch.exp(log_scales) * (torch.log(u) - torch.log(1. - u))
    x0 = torch.clamp(torch.clamp(x[:, :, :, 0], min=-1.), max=1.)
    x1 = torch.clamp(torch.clamp(
        x[:, :, :, 1] + coeffs[:, :, :, 0] * x0, min=-1.), max=1.)
    x2 = torch.clamp(torch.clamp(
        x[:, :, :, 2] + coeffs[:, :, :, 1] * x0 + coeffs[:, :, :, 2] * x1, min=-1.), max=1.)

    out = torch.cat([x0.view(xs[:-1] + [1]), x1.view(xs[:-1] + [1]), x2.view(xs[:-1] + [1])], dim=3)
    # put back in Pytorch ordering
    out = out.permute(0, 3, 1, 2)
    return out



import json
import os
import pickle

from booster.utils import available_device
from torch.distributions import Bernoulli

from .discretized_mixture_logits import DiscretizedMixtureLogits
from .logging import load_model
from ..datasets import get_binmnist_datasets, get_cifar10_datasets
from ..evaluation import VariationalInference
from ..model.deepvae import DeepVae


def restore_session(logdir, device='auto'):
    """load model from a saved session"""

    if logdir[-1] == '/':
        logdir = logdir[:-1]
    run_id = logdir.split('/')[-1]

    # load the hyperparameters and arguments
    hyperparameters = pickle.load(open(os.path.join(logdir, "hyperparameters.p"), "rb"))
    opt = json.load(open(os.path.join(logdir, "config.json")))

    # instantiate the model
    model = DeepVae(**hyperparameters)
    device = available_device() if device == 'auto' else device
    model.to(device)

    # load pretrained weights
    load_model(model, logdir)

    # define likelihood and evaluator
    likelihood = {'cifar': DiscretizedMixtureLogits(opt['nr_mix']), 'binmnist': Bernoulli}[opt['dataset']]
    evaluator = VariationalInference(likelihood, iw_samples=1)

    # load the dataset
    if opt['dataset'] == 'binmnist':
        train_dataset, valid_dataset, test_dataset = get_binmnist_datasets(opt['data_root'])
    elif opt['dataset'] == 'cifar10':
        from torchvision.transforms import Lambda

        transform = Lambda(lambda x: x * 2 - 1)
        train_dataset, valid_dataset, test_dataset = get_cifar10_datasets(opt.data_root, transform=transform)
    else:
        raise NotImplementedError

    return {
        'model': model,
        'device': device,
        'run_id': run_id,
        'hyperparameters': hyperparameters,
        'opt': hyperparameters,
        'likelihood': likelihood,
        'evaluator': evaluator,
        'train_dataset': train_dataset,
        'valid_dataset': valid_dataset,
        'test_dataset': test_dataset,
    }



import os
import pickle as pkl
from urllib.request import urlretrieve

import numpy as np
import torch
from torch.utils.data import Dataset


def load_mnist_binarized(root):
    datapath = os.path.join(root, 'bin-mnist')
    if not os.path.exists(datapath):
        os.makedirs(datapath)
    dataset = os.path.join(datapath, "mnist.pkl.gz")

    if not os.path.isfile(dataset):

        datafiles = {
            "train": "http://www.cs.toronto.edu/~larocheh/public/"
                     "datasets/binarized_mnist/binarized_mnist_train.amat",
            "valid": "http://www.cs.toronto.edu/~larocheh/public/datasets/"
                     "binarized_mnist/binarized_mnist_valid.amat",
            "test": "http://www.cs.toronto.edu/~larocheh/public/datasets/"
                    "binarized_mnist/binarized_mnist_test.amat"
        }
        datasplits = {}
        for split in datafiles.keys():
            print("Downloading %s data..." % (split))
            datasplits[split] = np.loadtxt(urlretrieve(datafiles[split])[0])

        pkl.dump([datasplits['train'], datasplits['valid'], datasplits['test']], open(dataset, "wb"))

    x_train, x_valid, x_test = pkl.load(open(dataset, "rb"))
    return x_train, x_valid, x_test


class BinMNIST(Dataset):
    """Binary MNIST dataset"""

    def __init__(self, data, device='cpu', transform=None):
        h, w, c = 28, 28, 1
        self.device = device
        self.data = torch.tensor(data, dtype=torch.float).view(-1, c, h, w)
        self.transform = transform

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        sample = self.data[idx]
        if self.transform:
            sample = self.transform(sample)
        return sample.to(self.device)


def get_binmnist_datasets(root, device='cpu'):
    x_train, x_valid, x_test = load_mnist_binarized(root)
    x_train = np.append(x_train, x_valid, axis=0)  # https://github.com/casperkaae/LVAE/blob/master/run_models.py (line 401)
    return BinMNIST(x_train, device=device), BinMNIST(x_test, device=device), BinMNIST(x_test, device=device)



import os
import pickle as pkl
import tarfile
from urllib.request import urlretrieve

import numpy as np
import torch
from torch.utils.data import Dataset

def quantisize(images, levels):
    return (np.digitize(images, np.arange(levels) / levels) - 1).astype('i')


def load_cifar(root, levels=256, with_y=False):
    dataset = 'cifar-10-python.tar.gz'
    data_dir, data_file = os.path.split(dataset)
    if data_dir == "" and not os.path.isfile(dataset):
        # Check if dataset is in the data directory.
        new_path = os.path.join(root, dataset)
        if os.path.isfile(new_path) or data_file == 'cifar-10-python.tar.gz':
            dataset = new_path

    if (not os.path.isfile(dataset)) and data_file == 'cifar-10-python.tar.gz':
        origin = (
            'http://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz'
        )
        print("Downloading data from {}...".format(origin))
        urlretrieve(origin, dataset)

    f = tarfile.open(dataset, 'r:gz')
    b1 = pkl.load(f.extractfile("cifar-10-batches-py/data_batch_1"), encoding="bytes")
    b2 = pkl.load(f.extractfile("cifar-10-batches-py/data_batch_2"), encoding="bytes")
    b3 = pkl.load(f.extractfile("cifar-10-batches-py/data_batch_3"), encoding="bytes")
    b4 = pkl.load(f.extractfile("cifar-10-batches-py/data_batch_4"), encoding="bytes")
    b5 = pkl.load(f.extractfile("cifar-10-batches-py/data_batch_5"), encoding="bytes")
    test = pkl.load(f.extractfile("cifar-10-batches-py/test_batch"), encoding="bytes")
    train_x = np.concatenate([b1[b'data'], b2[b'data'], b3[b'data'], b4[b'data'], b5[b'data']], axis=0) / 255.
    train_x = np.asarray(train_x, dtype='float32')
    train_t = np.concatenate([np.array(b1[b'labels']),
                              np.array(b2[b'labels']),
                              np.array(b3[b'labels']),
                              np.array(b4[b'labels']),
                              np.array(b5[b'labels'])], axis=0)

    test_x = test[b'data'] / 255.
    test_x = np.asarray(test_x, dtype='float32')
    test_t = np.array(test[b'labels'])
    f.close()

    train_x = train_x.reshape((train_x.shape[0], 3, 32, 32))
    test_x = test_x.reshape((test_x.shape[0], 3, 32, 32))
    train_x = quantisize(train_x, levels) / (levels - 1.)
    test_x = quantisize(test_x, levels) / (levels - 1.)

    if with_y:
        return (train_x, train_t), (test_x, test_t)
    return train_x, test_x


class Cifar10Dataset(Dataset):
    """Binary MNIST dataset"""

    def __init__(self, data, device='cpu', transform=None):
        self.device = device
        self.data = torch.tensor(data, dtype=torch.float)
        self.transform = transform

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        sample = self.data[idx]
        if self.transform:
            sample = self.transform(sample)
        return sample.to(self.device)


def get_cifar10_datasets(root, levels=256, **kwargs):
    x_train, x_test = load_cifar(root, levels=levels)
    return Cifar10Dataset(x_train, **kwargs), Cifar10Dataset(x_test, **kwargs), Cifar10Dataset(x_test, **kwargs)


from time import time

import torch


def append_ellapsed_time(func):
    def wrapper(*args, **kwargs):
        start_time = time()
        diagnostics = func(*args, **kwargs)
        diagnostics['info']['elapsed-time'] = time() - start_time
        return diagnostics

    return wrapper


@append_ellapsed_time
def training_step(x, model, evaluator, optimizer, scheduler=None, **kwargs):
    optimizer.zero_grad()
    model.train()

    loss, diagnostics, output = evaluator(model, x, **kwargs)
    loss = loss.mean(0)

    loss.backward()
    optimizer.step()
    if scheduler is not None:
        scheduler.step()

    return diagnostics


@torch.no_grad()
@append_ellapsed_time
def test_step(x, model, evaluator, **kwargs):
    model.eval()

    loss, diagnostics, output = evaluator(model, x, **kwargs)

    return diagnostics



import warnings

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn


def getSAMEPadding(tensor_shape, conv):
    """
    return the padding to apply to a given convolution such as it reproduces the 'SAME' behavior from Tensorflow
    This works as well for pooling layers.
    args:
        tensor_shape (tuple): input tensor shape (B x C x D)
        conv (nn.Module): convolution object
    returns:
        sym_padding, unsym_padding: symmetric and unsymmetric padding to apply,

    """
    tensor_shape = np.asarray(tensor_shape)[2:]
    kernel_size = np.asarray(conv.kernel_size)
    pad = np.asarray(conv.padding)
    dilation = np.asarray(conv.dilation) if hasattr(conv, 'dilation') else 1
    stride = np.asarray(conv.stride)
    # handle pooling layers
    if not hasattr(conv, 'transposed'):
        conv.transposed = False
    else:
        assert len(tensor_shape) == len(kernel_size), "tensor is not the same dimension as the kernel"
    if not conv.transposed:
        effective_filter_size = (kernel_size - 1) * dilation + 1
        output_size = (tensor_shape + stride - 1) // stride
        padding_input = np.maximum(0, (output_size - 1) * stride + (kernel_size - 1) * dilation + 1 - tensor_shape)
        odd_padding = (padding_input % 2 != 0)
        sym_padding = tuple(padding_input // 2)
        unsym_padding = [y for x in odd_padding for y in [0, int(x)]]
    else:
        padding_input = kernel_size - stride
        sym_padding = None
        unsym_padding = [y for x in padding_input for y in
                         [-int(np.floor(int(x) / 2)), -int(np.floor(int(x) / 2) + int(x) % 2)]]
    return sym_padding, unsym_padding


class PaddedConv(nn.Module):
    """
    wrapps convolution instance with SAME padding (as in tensorflow)
    This requires providing the input tensor shape.
    """

    def __init__(self, tensor_shape, conv):
        """
        args:
            tensor_shape (tuple): input tensor shape
            conv (nn.Module): convolution instance
        """
        super(PaddedConv, self).__init__()
        # input and output shapes
        self._input_shp = tensor_shape
        self._output_shp = np.asarray(tensor_shape)
        stride_factor = np.asarray(conv.stride)
        self._output_shp[2:] = self._output_shp[2:] // stride_factor if (
                not hasattr(conv, 'transposed') or not conv.transposed) else self._output_shp[2:] * stride_factor
        self._output_shp[1] = conv.out_channels if hasattr(conv, 'out_channels') else self._output_shp[1]
        self._output_shp = tuple(self._output_shp.astype(int))
        # get paddings
        sym_padding, self.unsym_padding = getSAMEPadding(tensor_shape, conv)
        if not conv.transposed:
            conv.padding = sym_padding
        self.conv = conv

    def forward(self, x):
        x = F.pad(x, self.unsym_padding) if not self.conv.transposed else x
        x = self.conv(x)
        x = F.pad(x, self.unsym_padding) if self.conv.transposed else x
        return x

    def init_parameters(self, x):
        x = F.pad(x, self.unsym_padding) if not self.conv.transposed else x
        self.conv.init_parameters(x)

    @property
    def input_shape(self):
        return self._input_shp

    @property
    def output_shape(self):
        return self._output_shp


class PaddedNormedConv(nn.Module):
    """
    A class to handle both normalization and SAME padding (as in tensorflow) for convolutions.
    This class also handles data dependent initialization for weight normalization
    """

    def __init__(self, tensor_shape, conv, weightnorm=True):
        """
        args:
            tensor_shape (tuple): input tensor shape (B x C x D)
            conv (nn.Module): convolution instance of type Conv1d, ConvTranspose1d, Conv2d or ConvTranspose2d
            weightnorm (bool): use weight normalization
        """
        super(PaddedNormedConv, self).__init__()
        self.register_buffer("initialized", torch.tensor(False))

        # paddding
        self._input_shp = tensor_shape
        self._output_shp = np.asarray(tensor_shape)
        stride_factor = np.asarray(conv.stride)
        self._output_shp[2:] = self._output_shp[2:] // stride_factor if (
                not hasattr(conv, 'transposed') or not conv.transposed) else self._output_shp[2:] * stride_factor
        self._output_shp[1] = conv.out_channels if hasattr(conv, 'out_channels') else self._output_shp[1]
        self._output_shp = tuple(self._output_shp.astype(int))

        # get paddings
        sym_padding, self.unsym_padding = getSAMEPadding(tensor_shape, conv)
        if not conv.transposed:
            conv.padding = sym_padding
        self.conv = conv

        # add batch norm
        if not weightnorm:
            self.weightnorm = False
        else:
            self.weightnorm = True
            dim = 1 if self.conv.transposed else 0
            self.conv = nn.utils.weight_norm(self.conv, dim=dim, name="weight")

    def forward(self, x):
        x = F.pad(x, self.unsym_padding) if not self.conv.transposed else x
        if not self.initialized:
            self.init_parameters(x)
        x = self.conv(x)

        x = F.pad(x, self.unsym_padding) if self.conv.transposed else x
        return x

    @property
    def input_shape(self):
        return self._input_shp

    @property
    def output_shape(self):
        return self._output_shp

    def init_parameters(self, x, init_scale=0.05, eps=1e-8):
        self.initialized = True + self.initialized
        if self.weightnorm:
            # initial values
            self.conv._parameters['weight_v'].data.normal_(mean=0, std=init_scale)
            self.conv._parameters['weight_g'].data.fill_(1.)
            self.conv._parameters['bias'].data.fill_(0.)
            init_scale = .01
            # data dependent init
            x = self.conv(x)
            t = x.view(x.size()[0], x.size()[1], -1)
            t = t.permute(0, 2, 1).contiguous()
            t = t.view(-1, t.size()[-1])
            m_init, v_init = torch.mean(t, 0), torch.var(t, 0)
            scale_init = init_scale / torch.sqrt(v_init + eps)
            if self.conv.transposed:
                self.conv._parameters['weight_g'].data = self.conv._parameters['weight_g'].data * scale_init[None,
                                                                                                  :].view(
                    self.conv._parameters['weight_g'].data.size())
                self.conv._parameters['bias'].data = self.conv._parameters['bias'].data - m_init * scale_init
            else:
                self.conv._parameters['weight_g'].data = self.conv._parameters['weight_g'].data * scale_init[:,
                                                                                                  None].view(
                    self.conv._parameters['weight_g'].data.size())
                self.conv._parameters['bias'].data = self.conv._parameters['bias'].data - m_init * scale_init
            return scale_init[None, :, None, None] * (x - m_init[None, :, None, None]) if len(
                self._input_shp) > 3 else scale_init[None, :, None] * (x - m_init[None, :, None])


class ResidualConnection(nn.Module):
    """
    Handles residual connections for tensors with different shapes.
    Apply padding and/or avg pooling to the input when necessary
    """

    def __init__(self, input_shape, output_shape, residual=True):
        """
        args:
            input_shape (tuple): input module shape x
            output_shape (tuple): output module shape y=f(x)
            residual (bool): apply residual conenction y' = y+x = f(x)+x
        """
        super().__init__()
        self.residual = residual
        self.input_shape = input_shape
        self.output_shape = output_shape
        is_text = len(input_shape) == 3

        # residual: features
        if residual and self.output_shape[1] < self.input_shape[1]:
            pad = int(self.output_shape[1]) - int(self.input_shape[1])
            self.redidual_padding = [0, 0, 0, pad] if is_text else [0, 0, 0, 0, 0, pad]

        elif residual and self.output_shape[1] > self.input_shape[1]:
            pad = int(self.output_shape[1]) - int(self.input_shape[1])
            self.redidual_padding = [0, 0, 0, pad] if is_text else [0, 0, 0, 0, 0, pad]
            warnings.warn("The input has more feature maps than the output. There will be no residual connection for this layer.")
            self.residual = False
        else:
            self.redidual_padding = None

        # residual: dimension
        if residual and list(output_shape)[2:] < list(input_shape)[2:]:
            pool_obj = nn.AvgPool1d if len(output_shape[2:]) == 1 else nn.AvgPool2d
            stride = tuple((np.asarray(input_shape)[2:] // np.asarray(output_shape)[2:]).tolist())
            self.residual_op = PaddedConv(input_shape, pool_obj(3, stride=stride))

        elif residual and list(output_shape)[2:] > list(input_shape)[2:]:
            warnings.warn(
                "The height and width of the output are larger than the input. There will be no residual connection for this layer.")
            # self.residual_op = nn.UpsamplingBilinear2d(size=self.output_shape[2:])
            self.residual = False
        else:
            self.residual_op = None

    def forward(self, y, x):
        if self.residual:
            x = F.pad(x, self.redidual_padding) if self.redidual_padding is not None else x
            x = self.residual_op(x) if self.residual_op is not None else x
            y = y + x
        return y


def getConvolutionOutputShape(tensor_shape, conv):
    """
    compute the output shape of a convolution given the input tensor shape
    args:
        tensor_shape (tuple): input tensor shape (B x C x D)
        conv (nn.Module): convolution object
    returns:
        output_shape (tuple): expected output shape
    """
    assert tensor_shape[1] == conv.in_channels, "tensor and kernel do not have the same nuæber of features"
    tensor_shape = np.asarray(tensor_shape)
    kernel_size = np.asarray(conv.kernel_size)
    pad = np.asarray(conv.padding)
    dilation = np.asarray(conv.dilation)
    stride = np.asarray(conv.stride)
    assert len(tensor_shape) - 2 == len(kernel_size), "tensor is not the same dimension as the kernel"
    dims = tensor_shape[2:]
    if not conv.transposed:
        out_dims = ((dims + (2 * pad) - (dilation * (kernel_size - 1)) - 1) / stride) + 1
    else:
        out_padding = np.asarray(conv.output_padding)
        out_dims = dilation * ((dims - 1) * stride - 2 * padding + kernel_size) + out_padding
    out_dims = np.floor(out_dims).astype(int)
    return (tensor_shape[0], conv.out_channels, *out_dims)


if __name__ == '__main__':
    # test getConvolutionOutputShape
    for x in [torch.zeros(1, 3, 33, 33), torch.zeros(1, 3, 64, 64)]:
        for kernel_size in [3, 5, 6, 8]:
            for padding in [1, 2]:
                for stride in [1, 2, 3]:
                    for dilation in [1]:
                        for Conv in [nn.Conv2d, nn.ConvTranspose2d]:
                            conv = Conv(3, 18, kernel_size=kernel_size, padding=padding, stride=stride,
                                        dilation=dilation)
                            x_shp = tuple(x.size())
                            c_shp = getConvolutionOutputShape(x_shp, conv)
                            true_shp = tuple(conv(x).size())
                            assert c_shp[2:] == true_shp[2:]

    # test getPadding
    for x in [torch.zeros(1, 3, 33, 32), torch.zeros(1, 3, 15, 15)]:
        for Conv in [nn.ConvTranspose2d, nn.Conv2d]:
            for stride in [2, 1]:
                for dilation in [1, 2]:
                    for kernel_size in [3, 5, 6]:
                        conv = Conv(3, 5, kernel_size=kernel_size, stride=stride, dilation=dilation)
                        # print("kernel_size:",kernel_size, " stride:", stride," trans.:",conv.transposed, " x:", x.size())
                        x_shp = tuple(x.size())
                        padding, odd_padding = getSAMEPadding(x_shp, conv)
                        if not conv.transposed:
                            conv.padding = padding
                        x_shp = x_shp[2:]
                        expected_shp = tuple([t * stride for t in x_shp]) if conv.transposed else tuple(
                            [t // stride for t in x_shp])
                        y = F.pad(x, odd_padding) if not conv.transposed else x
                        y = conv(y)
                        y = F.pad(y, odd_padding) if conv.transposed else y
                        true_shp = tuple(y.size())[2:]
                        # print(true_shp,expected_shp,np.asarray(true_shp)-np.asarray(expected_shp),kernel_size-stride,'\n')
                        assert true_shp >= expected_shp

    # test SAMEpaddingConv
    for x in [torch.zeros(1, 3, 33, 32), torch.zeros(1, 3, 15, 15)]:
        for Conv in [nn.ConvTranspose2d, nn.Conv2d]:
            for stride in [2, 1]:
                for dilation in [1, 2]:
                    for kernel_size in [3, 5, 6]:
                        conv = Conv(3, 5, kernel_size=kernel_size, stride=stride, dilation=dilation)
                        # print("kernel_size:",kernel_size, " stride:", stride," trans.:",conv.transposed, " x:", x.size())
                        x_shp = tuple(x.size())
                        conv = SAMEpaddingConv(x_shp, conv)
                        x_shp = x_shp[2:]
                        expected_shp = tuple([t * stride for t in x_shp]) if conv.conv.transposed else tuple(
                            [t // stride for t in x_shp])
                        y = conv(x)
                        true_shp = tuple(y.size())[2:]
                        # print(true_shp,expected_shp,np.asarray(true_shp)-np.asarray(expected_shp),kernel_size-stride,'\n')
                        assert true_shp >= expected_shp

    # test SAMEpaddingConv
    for x in [torch.zeros(1, 3, 33, 32), torch.zeros(1, 3, 15, 15)]:
        for Conv in [nn.MaxPool2d]:
            for stride in [2, 1]:
                for dilation in [1]:
                    for kernel_size in [3, 5]:
                        conv = Conv(kernel_size=kernel_size, stride=stride, dilation=dilation)
                        # print("kernel_size:",kernel_size, " stride:", stride," trans.:",conv.transposed, " x:", x.size())
                        x_shp = tuple(x.size())
                        conv = SAMEpaddingConv(x_shp, conv)
                        x_shp = x_shp[2:]
                        expected_shp = tuple([t * stride for t in x_shp]) if conv.conv.transposed else tuple(
                            [t // stride for t in x_shp])
                        y = conv(x)
                        true_shp = tuple(y.size())[2:]
                        # print(true_shp,expected_shp,np.asarray(true_shp)-np.asarray(expected_shp),kernel_size-stride,'\n')
                        assert true_shp >= expected_shp



from typing import *

from .convolution import *
from .linear import *


class GatedResNet(nn.Module):
    def __init__(self, input_shape: Tuple, dim: Tuple, aux_shape: Optional[Tuple] = None, weightnorm: bool = True,
                 act: nn.Module = nn.ReLU,
                 transposed: bool = False, dropout: Optional[float] = None, residual: bool = True, **kwargs: Any):
        """
        args:
            input_shape (tuple): input tensor shape (B x C x *D)
            dim (tuple): convolution dimensions (filters, kernel_size, stride)
            aux_shape (tuple): auxiliary input tensor shape (B x C x *D). None means no auxialiary input
            weightnorm (bool): use weight normalization
            act (nn.Module): activation function class
            transposed (bool): transposed or not
            dropout (float): dropout value. None is no dropout
            residual (bool): use residual connections
        """
        super(GatedResNet, self).__init__()

        # define convolution and transposed convolution objects
        if len(input_shape[2:]) == 1:
            conv_obj = nn.Conv1d
            deconv_obj = nn.ConvTranspose1d
            dropout_obj = nn.Dropout
        else:
            conv_obj = nn.Conv2d
            deconv_obj = nn.ConvTranspose2d
            dropout_obj = nn.Dropout

        # some parameters
        C_in = input_shape[1]
        self.residual = residual
        self.transposed = transposed
        self.act = act()

        # conv 1
        conv1 = conv_obj(C_in, dim[0], dim[1], 1)
        self.conv1 = PaddedNormedConv(input_shape, conv1, weightnorm=weightnorm)
        shp = self.conv1.output_shape

        # dropout
        self.dropout = dropout_obj(dropout) if dropout is not None else dropout

        # conv 2
        if self.transposed and dim[2] > 1:
            conv2 = deconv_obj(dim[0], 2 * dim[0], dim[1], dim[2])
        else:
            conv2 = conv_obj(dim[0], 2 * dim[0], dim[1], dim[2])

        self.conv2 = PaddedNormedConv(shp, conv2, weightnorm=weightnorm)

        # input / output shapes
        shp = list(self.conv2.output_shape)
        shp[1] = shp[1] // 2  # gated
        self._input_shape = input_shape
        self._output_shape = tuple(shp)
        self.aux_shape = aux_shape

        # residual connections
        self.residual_op = ResidualConnection(self._input_shape, shp, residual)

        # aux op
        if aux_shape is not None:
            if list(aux_shape[2:]) > list(input_shape[2:]):
                stride = tuple(np.asarray(aux_shape[2:]) // np.asarray(input_shape[2:]))
                aux_conv = conv_obj(aux_shape[1], dim[0], dim[1], stride)
                self.aux_op = PaddedNormedConv(aux_shape, aux_conv, weightnorm=weightnorm)

            elif list(aux_shape[2:]) < list(input_shape[2:]):
                stride = tuple(np.asarray(input_shape[2:]) // np.asarray(aux_shape[2:]))
                aux_conv = deconv_obj(aux_shape[1], dim[0], dim[1], stride)
                self.aux_op = PaddedNormedConv(aux_shape, aux_conv, weightnorm=weightnorm)

            else:
                aux_conv = conv_obj(aux_shape[1], dim[0], 1, 1)  # conv with kernel 1
                self.aux_op = PaddedNormedConv(aux_shape, aux_conv, weightnorm=weightnorm)

        else:
            self.aux_op = None

    def forward(self, x: torch.Tensor, aux: Optional[torch.Tensor] = None, **kwargs: Any) -> torch.Tensor:

        # input activation: x = act(x)
        x_act = self.act(x) if self.residual else x

        # conv 1: y = conv(x)
        y = self.conv1(x_act)

        # merge aux with x: y = y + f(aux)
        y = y + self.aux_op(self.act(aux)) if self.aux_op is not None else y

        # y = act(y)
        y = self.act(y)

        # dropout
        y = self.dropout(y) if self.dropout else y

        # conv 2: y = conv(y)
        y = self.conv2(y)

        # gate: y = y_1 * sigmoid(y_2)
        h_stack1, h_stack2 = y.chunk(2, 1)
        sigmoid_out = torch.sigmoid(h_stack2)
        y = (h_stack1 * sigmoid_out)

        # resiudal connection: y = y + x
        y = self.residual_op(y, x)

        return y

    @property
    def input_shape(self) -> Tuple:
        return self._input_shape

    @property
    def output_shape(self) -> Tuple:
        return self._output_shape


class ResMLP(nn.Module):
    def __init__(self, input_shape: Tuple[int], dim: int, aux_shape: Optional[int] = None, weightnorm: bool = True,
                 act: nn.Module = nn.ReLU, transposed: bool = False, dropout: float = None, residual: bool = True,
                 mlp_layers: int = 1,
                 **kwargs: Any):
        super().__init__()

        # convert parameters
        ninp = input_shape[1]
        nhid = dim
        naux = aux_shape[1] if aux_shape is not None else 0
        nlayers = mlp_layers

        # params
        self._input_shape = input_shape
        self._output_shape = (-1, dim)

        # model
        self.act = act()
        self.ninp = ninp
        self.naux = naux
        if naux is None:
            naux = 0
        self.residual = NormedLinear(ninp, nhid, weightnorm) if ninp != nhid else None
        if residual:
            layers = [act(), NormedLinear(ninp + naux, nhid, weightnorm), nn.BatchNorm1d(nhid)]
        else:
            layers = [NormedLinear(ninp + naux, nhid, weightnorm), nn.BatchNorm1d(nhid)]
        layers += (nlayers - 1) * [act(), NormedLinear(nhid, nhid, weightnorm), nn.BatchNorm1d(nhid)]
        self.layers = nn.Sequential(*layers)

    def forward(self, x, aux=None):
        if self.residual is not None:
            r = self.residual(x)
        else:
            r = x
        if aux is not None:
            x = torch.cat([x, aux], 1)
        return r + self.layers(x)

    @property
    def input_shape(self):
        return self._input_shape

    @property
    def output_shape(self):
        return self._output_shape



from typing import *

import torch
from torch import nn

from .architectures import get_deep_vae_mnist
from .stage import VaeStage, LvaeStage, BivaStage
from .utils import DataCollector
from ..layers import PaddedNormedConv


class DeepVae(nn.Module):
    """
    A Deep Hierarchical VAE.
    The model is a stack of N stages. Each stage features an inference and a generative path.
    Depending on the choice of the stage, multiple models can be implemented:
    - VAE: https://arxiv.org/abs/1312.6114
    - LVAE: https://arxiv.org/abs/1602.02282
    - BIVA: https://arxiv.org/abs/1902.02102
    """

    def __init__(self,
                 Stage: Any = BivaStage,
                 tensor_shp: Tuple[int] = (-1, 1, 28, 28),
                 padded_shp: Optional[Tuple] = None,
                 stages: List[List[Tuple]] = None,
                 latents: List = None,
                 nonlinearity: str = 'elu',
                 q_dropout: float = 0.,
                 p_dropout: float = 0.,
                 features_out: Optional[int] = None,
                 lambda_init: Optional[Callable] = None,
                 projection: Optional[nn.Module] = None,
                 **kwargs):

        """
        Initialize the Deep VAE model.
        :param Stage: stage constructor (VaeStage, LvaeStage, BivaStage)
        :param tensor_shp: Input tensor shape (batch_size, channels, *dimensions)
        :param padded_shp: pad input tensor to this shape
        :param stages: a list of list of tuple, each tuple describing a convolutional block (filters, stride, kernel_size)
        :param latents: a list describing the stochastic layers for each stage
        :param nonlinearity: activation function (gelu, elu, relu, tanh)
        :param q_dropout: inference dropout value
        :param p_dropout: generative dropout value
        :param features_out: optional number of output features if different from the input
        :param lambda_init: lambda function applied to the input
        :param projection: projection layer with constructor __init__(output_shape)

        :param kwargs: additional arugments passed to each stage
        """
        super().__init__()
        stages, latents = self.get_default_architecture(stages, latents)

        self.input_tensor_shape = tensor_shp
        self.lambda_init = lambda_init

        # input padding
        if padded_shp is not None:
            padding = [[(t - o) // 2, (t - o) // 2] for t, o in zip(padded_shp, tensor_shp[2:])]
            self.pad = [u for pads in padding for u in pads]
            self.unpad = [-u for u in self.pad]
            in_shp = [*tensor_shp[:2], *padded_shp]
        else:
            self.pad = None
            in_shp = tensor_shp

        # select activation class
        Act = {'elu': nn.ELU, 'relu': nn.ReLU, 'tanh': nn.Tanh()}[nonlinearity]

        # initialize the inference path
        stages_ = []
        block_args = {'act': Act, 'q_dropout': q_dropout, 'p_dropout': p_dropout}

        input_shape = {'x': in_shp}
        for i, (conv_data, z_data) in enumerate(zip(stages, latents)):
            top = i == len(stages) - 1
            bottom = i == 0

            stage = Stage(input_shape, conv_data, z_data, top=top, bottom=bottom, **block_args, **kwargs)

            input_shape = stage.q_output_shape
            stages_ += [stage]

        self.stages = nn.ModuleList(stages_)

        if projection is None:
            # output convolution
            tensor_shp = self.stages[0].p_output_shape['d']
            if features_out is None:
                features_out = self.input_tensor_shape[1]
            conv_obj = nn.Conv2d if len(tensor_shp) == 4 else nn.Conv1d
            conv_out = conv_obj(tensor_shp[1], features_out, 1)
            conv_out = PaddedNormedConv(tensor_shp, conv_out, weightnorm=True)
            self.projection = nn.Sequential(Act(), conv_out)
        else:
            tensor_shp = self.stages[0].forward_shape['d']
            self.projection = projection(tensor_shp)

    def get_default_architecture(self, stages, latents):
        if stages is None:
            stages, _ = get_deep_vae_mnist()

        if latents is None:
            _, latens = get_deep_vae_mnist()

        return stages, latents

    def infer(self, x: torch.Tensor, **kwargs: Any) -> List[Dict]:
        """
        Forward pass through the inference network and return the posterior of each layer order from the top to the bottom.
        :param x: input tensor
        :param kwargs: additional arguments passed to each stage
        :return: a list that contains the data for each stage
        """
        posteriors = []
        data = {'x': x}
        for stage in self.stages:
            data, posterior = stage.infer(data, **kwargs)
            posteriors += [posterior]

        return posteriors

    def generate(self, posteriors: Optional[List], **kwargs) -> Dict[str, torch.Tensor]:
        """
        Forward pass through the generative model, compute KL and return reconstruction x_, KL and auxiliary data.
        If no posterior is provided, the prior is sampled.
        :param posteriors: a list containing the posterior for each stage
        :param kwargs: additional arguments passed to each stage
        :return: {'x_': reconstruction logits, 'kl': kl for each stage, **auxiliary}
        """
        if posteriors is None:
            posteriors = [None for _ in self.stages]

        output_data = DataCollector()
        x = {}
        for posterior, stage in zip(posteriors[::-1], self.stages[::-1]):
            x, data = stage(x, posterior, **kwargs)
            output_data.extend(data)

        # output convolution
        x = self.projection(x['d'])

        # undo padding
        if self.pad is not None:
            x = nn.functional.pad(x, self.unpad)

        # sort data: [z1, z2, ..., z_L]
        output_data = output_data.sort()

        return {'x_': x, **output_data}

    def forward(self, x: torch.Tensor, **kwargs: Any) -> Dict[str, torch.Tensor]:
        """
        Forward pass through the inference model, the generative model and compute KL for each stage.
        x_ = p_\theta(x|z), z \sim q_\phi(z|x)
        kl_i = log q_\phi(z_i | h) - log p_\theta(z_i | h)

        :param x: input tensor
        :param kwargs: additional arguments passed to each stage
        :return: {'x_': reconstruction logits, 'kl': kl for each stage, **auxiliary}
        """

        if self.pad is not None:
            x = nn.functional.pad(x, self.pad)

        if self.lambda_init is not None:
            x = self.lambda_init(x)

        posteriors = self.infer(x, **kwargs)

        data = self.generate(posteriors, N=x.size(0), **kwargs)

        return data

    def sample_from_prior(self, N: int, **kwargs: Any) -> Dict[str, torch.Tensor]:
        """
        Sample the prior and pass through the generative model.
        x_ = p_\theta(x|z), z \sim p_\theta(z)

        :param N: number of samples (batch size)
        :param kwargs: additional arguments passed to each stage
        :return: {'x_': sample logits}
        """
        return self.generate(None, N=N, **kwargs)


class BIVA(DeepVae):
    def __init__(self, **kwargs):
        kwargs.pop('Stage', None)
        super().__init__(Stage=BivaStage, **kwargs)


class LVAE(DeepVae):
    def __init__(self, **kwargs):
        kwargs.pop('Stage', None)
        super().__init__(Stage=LvaeStage, **kwargs)


class VAE(DeepVae):
    def __init__(self, **kwargs):
        kwargs.pop('Stage', None)
        super().__init__(Stage=VaeStage, **kwargs)



import numpy as np
import torch


class FreeBits():
    """
    free bits: https://arxiv.org/abs/1606.04934
    Assumes a each of the dimension to be one group
    """

    def __init__(self, min_KL: float):
        self.min_KL = min_KL

    def __call__(self, kls: torch.Tensor) -> torch.Tensor:
        """
        Apply freebits over tensor. The freebits budget is distributed equally among dimensions.
        The returned freebits KL is equal to max(kl, freebits_per_dim, dim = >0)
        :param kls: KL of shape [batch size, *dimensions]
        :return:  freebits KL of shape [batch size, *dimensions]
        """

        # equally divide freebits budget over the dimensions
        dimensions = np.prod(kls.shape[1:])
        min_KL_per_dim = self.min_KL / dimensions if len(kls.shape) > 1 else self.min_KL
        min_KL_per_dim = min_KL_per_dim * torch.ones_like(kls)

        # apply freebits
        freebits_kl = torch.cat([kls.unsqueeze(-1), min_KL_per_dim.unsqueeze(-1)], -1)
        freebits_kl = torch.max(freebits_kl, dim=-1)[0]

        return freebits_kl



from typing import *

import torch
from torch.optim.lr_scheduler import _LRScheduler


def shp_cat(shps: List[Tuple[int]], dim: int):
    """concatenate tensor shapes"""
    out = list(shps[0])
    out[dim] = sum(list(s)[dim] for s in shps)
    return tuple(out)


def detach_to_device(x, device):
    """detach, clone and or place on the right device"""
    if x is not None:
        if isinstance(x, torch.Tensor):
            return x.detach().clone().to(device)
        else:
            return torch.tensor(x, device=device, dtype=torch.float)
    else:
        return None


def batch_reduce(x, reduce=torch.sum):
    batch_size = x.size(0)
    return reduce(x.view(batch_size, -1), dim=-1)


def log_sum_exp(tensor, dim=-1, sum_op=torch.sum, eps: float = 1e-12, keepdim=False):
    """
    Uses the LogSumExp (LSE) as an approximation for the sum in a log-domain.
    :param tensor: Tensor to compute LSE over
    :param dim: dimension to perform operation over
    :param sum_op: reductive operation to be applied, e.g. torch.sum or torch.mean
    :return: LSE
    """
    max, _ = torch.max(tensor, dim=dim, keepdim=keepdim)
    return torch.log(sum_op(torch.exp(tensor - max), dim=dim, keepdim=keepdim) + eps) + max


class LowerBoundedExponentialLR(_LRScheduler):
    """Set the learning rate of each parameter group to the initial lr decayed
    by gamma every epoch. When last_epoch=-1, sets initial lr as lr.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        gamma (float): Multiplicative factor of learning rate decay.
        lower_bound (float): lower bound for the learning rate.
        last_epoch (int): The index of last epoch. Default: -1.
    """

    def __init__(self, optimizer, gamma, lower_bound, last_epoch=-1):
        self.gamma = gamma
        self.lower_bound = lower_bound
        super(LowerBoundedExponentialLR, self).__init__(optimizer, last_epoch)

    def _get_lr(self, base_lr):
        lr = base_lr * self.gamma ** self.last_epoch
        if lr < self.lower_bound:
            lr = self.lower_bound
        return lr

    def get_lr(self):
        return [self._get_lr(base_lr)
                for base_lr in self.base_lrs]



from collections import defaultdict
from typing import *

import torch


class DataCollector(defaultdict):
    def __init__(self):
        super().__init__(list)

    def extend(self, data: Dict[str, List[Optional[torch.Tensor]]]) -> None:
        """Append new data item"""
        for key, d in data.items():
            self[key] += d

    def sort(self) -> Dict[str, List[Optional[torch.Tensor]]]:
        """sort data and return"""
        for key, d in self.items():
            d = d[::-1]
            self[key] = [t for t in d if t is not None]

        return self



from typing import *

import torch
from torch import nn, Tensor
from torch.distributions import Normal

from ..layers import PaddedNormedConv, NormedDense
from ..utils import batch_reduce


class StochasticLayer(nn.Module):
    """
    An abstract class of a VAE stochastic layer.
    """

    def __init__(self, data: Dict, tensor_shp: Tuple[int], **kwargs: Any):
        super().__init__()
        self._output_shape = None
        self._input_shape = tensor_shp

    def forward(self, x: Optional[Tensor], inference: bool, sample: bool = True, N: Optional[int] = None, **kwargs) -> \
            Tuple[
                Tensor, Dict[str, Any]]:
        """
        Returns the distribution parametrized by x and sample if `sample1`=True. If no hidden state is provided, uses the prior.
        :param x: hidden state used to computed logits (Optional : None means using the prior)
        :param inference: inference mode
        :param sample: sample layer
        :param N: number of samples (when sampling from prior)
        :param kwargs: additional args passed ot the stochastic layer
        :return: (projected sample, data)
        """
        raise NotImplementedError

    def loss(self, q_data: Dict[str, Any], p_data: Dict[str, Any], **kwargs: Any) -> Dict[str, List[Any]]:
        """
        Compute the KL divergence and other auxiliary losses if required
        :param q_data: data received from the posterior forward pass
        :param p_data: data received from the prior forward pass
        :param kwargs: other parameters passed to the kl function
        :return: dictionary of losses {'kl': [values], 'auxiliary' : [aux_values], ...}
        """
        raise NotImplementedError

    @property
    def output_shape(self):
        return self._output_shape

    @property
    def input_shape(self):
        return self._input_shape


class DenseNormal(StochasticLayer):
    """
    A Normal stochastic layer parametrized by dense layers.
    """

    def __init__(self, data: Dict, tensor_shp: Tuple[int], top: bool = False, act: nn.Module = nn.ELU,
                 weightnorm: bool = True, log_var_act: Optional[Callable] = nn.Softplus, **kwargs):
        super().__init__(data, tensor_shp)

        self._input_shape = tensor_shp

        self.eps = 1e-8
        self.nz = data.get('N')
        self.tensor_shp = tensor_shp
        self.dim = 2
        self.act = act()
        self.log_var_act = log_var_act() if log_var_act is not None else None

        # stochastic layer and prior
        if top:
            prior = torch.zeros((2 * self.nz))
            self.register_buffer('prior', prior)

        # computes logits
        nz_in = 2 * self.nz
        self.qx2z = NormedDense(tensor_shp, nz_in, weightnorm=weightnorm)
        if not top:
            self.px2z = NormedDense(tensor_shp, nz_in, weightnorm=weightnorm)

        self._output_shape = (-1, self.nz)

    @property
    def output_shape(self):
        return self._output_shape

    @property
    def input_shape(self):
        return self._input_shape

    def compute_logits(self, x: Tensor, inference: bool) -> Tuple[Tensor, Tensor]:
        """
        Compute the logits of the distribution.
        :param x: input tensor
        :param inference: inference mode
        :return: logits
        """
        x = self.act(x)

        if inference:
            logits = self.qx2z(x)
        else:
            logits = self.px2z(x)

        # apply activation to logvar
        mu, logvar = logits.chunk(2, dim=1)
        if self.log_var_act is not None:
            logvar = self.log_var_act(logvar)
        return mu, logvar

    def forward(self, x: Optional[Tensor], inference: bool, sample: bool = True, N: Optional[int] = None, **kwargs) -> \
            Tuple[Tensor, Dict[str, Any]]:

        if x is None:
            mu, logvar = self.prior.expand(N, *self.prior.shape).chunk(2, dim=1)
        else:
            mu, logvar = self.compute_logits(x, inference)

        # sample layer
        std = logvar.mul(0.5).exp()
        dist = Normal(mu, std)

        z = dist.rsample() if sample else None

        return z, {'z': z, 'dist': dist}

    def loss(self, q_data: Dict[str, Any], p_data: Dict[str, Any], **kwargs: Any) -> Dict[str, List]:
        z_q = q_data.get('z')
        q = q_data.get('dist')
        p = p_data.get('dist')

        kl = q.log_prob(z_q) - p.log_prob(z_q)
        kl = batch_reduce(kl)

        return {'kl': [kl]}


class ConvNormal(StochasticLayer):
    """
    A Normal stochastic layer parametrized by convolutions.
    """

    def __init__(self, data: Dict, tensor_shp: Tuple[int], top: bool = False, act: nn.Module = nn.ELU,
                 learn_prior: bool = False, weightnorm: bool = True, log_var_act: Optional[Callable] = nn.Softplus,
                 **kwargs):
        super().__init__(data, tensor_shp)

        self.eps = 1e-8
        nhid = tensor_shp[1]
        self.nz = data.get('N')
        kernel_size = data.get('kernel')
        self.tensor_shp = tensor_shp
        self.input_shp = tensor_shp
        self.act = act()
        self.log_var_act = log_var_act() if log_var_act is not None else None

        # prior
        if top:
            prior = torch.zeros((2 * self.nz, *tensor_shp[2:]))

            if learn_prior:
                self.prior = nn.Parameter(prior)
            else:
                self.register_buffer('prior', prior)

        # computes logits
        nz_in = 2 * self.nz
        self.qx2z = PaddedNormedConv(tensor_shp, nn.Conv2d(nhid, nz_in, kernel_size), weightnorm=weightnorm)
        if not top:
            self.px2z = PaddedNormedConv(tensor_shp, nn.Conv2d(nhid, nz_in, kernel_size), weightnorm=weightnorm)

        # compute output shape
        nz_out = self.nz
        out_shp = (-1, nz_out, *tensor_shp[2:])
        self._output_shape = out_shp
        self._input_shape = tensor_shp

    @property
    def output_shape(self):
        return self._output_shape

    @property
    def input_shape(self):
        return self._input_shape

    def compute_logits(self, x: Tensor, inference: bool) -> Tuple[Tensor, Tensor]:
        """
        Compute the logits of the distribution.
        :param x: input tensor
        :param inference: inference mode
        :return: logits
        """
        x = self.act(x)
        if inference:
            logits = self.qx2z(x)
        else:
            logits = self.px2z(x)

        # apply activation to logvar
        mu, logvar = logits.chunk(2, dim=1)
        if self.log_var_act is not None:
            logvar = self.log_var_act(logvar)
        return mu, logvar

    def expand_prior(self, batch_size: int):
        return self.prior.expand(batch_size, *self.prior.shape).chunk(2, dim=1)

    def forward(self, x: Optional[Tensor], inference: bool, sample: bool = True, N: Optional[int] = None, **kwargs) -> \
            Tuple[
                Tensor, Dict[str, Any]]:

        if x is None:
            mu, logvar = self.expand_prior(N)
        else:
            mu, logvar = self.compute_logits(x, inference)

        # sample layer
        std = logvar.mul(0.5).exp()
        dist = Normal(mu, std)

        z = dist.rsample() if sample else None

        return z, {'z': z, 'dist': dist}

    def loss(self, q_data: Dict[str, Any], p_data: Dict[str, Any], **kwargs: Any) -> Dict[str, List]:
        z_q = q_data.get('z')
        q = q_data.get('dist')
        p = p_data.get('dist')

        kl = q.log_prob(z_q) - p.log_prob(z_q)
        kl = batch_reduce(kl)

        return {'kl': [kl]}



import numpy as np
import torch
from torch import nn


class NormedLinear(nn.Module):
    """
    Linear layer with normalization
    """

    def __init__(self, in_features, out_features, dim=-1, weightnorm=True):
        super(NormedLinear, self).__init__()
        """
        args:
            in_features (in): number of input features
            out_features (int): number of output features
            dim (int): dimension to aply transformation to
            weightnorm (bool): use weight normalization
        """
        self.register_buffer("initialized", torch.tensor(False))
        self._in_features = in_features
        self._out_features = out_features
        self.dim = dim
        self.linear = nn.Linear(self._in_features, out_features)
        # add batch norm
        if not weightnorm:
            self.weightnorm = False
        else:
            self.weightnorm = True
            self.linear = nn.utils.weight_norm(self.linear, dim=0, name="weight")

    def forward(self, x):
        # reshape in
        shp = list(x.size())
        dim = self.dim if self.dim >= 0 else x.dim() + self.dim
        if dim < x.dim() - 1:
            x = x.view(*x.size()[:dim + 1], -1)
            x = x.transpose(-1, -2).contiguous()
            shp_2 = list(x.shape)
            x = x.view(-1, x.size(-1))
            permute = True
        else:
            x = x.view(-1, x.size(-1))
            permute = False
        # init and transform
        if not self.initialized:
            self.init_parameters(x)
        x = self.linear(x)
        # reshape out
        shp[dim] = self._out_features
        if permute:
            shp_2[-1] = self._out_features
            x = x.view(shp_2).transpose(-1, -2)
            x = x.view(shp)
        else:
            x = x.view(shp)
        return x

    @property
    def input_shape(self):
        return (-1, self._in_features)

    @property
    def output_shape(self):
        return (-1, self._out_features)

    def init_parameters(self, x, init_scale=0.05, eps=1e-8):
        if self.weightnorm:
            # initial values
            self.linear._parameters['weight_v'].data.normal_(mean=0, std=init_scale)
            self.linear._parameters['weight_g'].data.fill_(1.)
            self.linear._parameters['bias'].data.fill_(0.)
            init_scale = .01
            # data dependent init
            x = self.linear(x)
            m_init, v_init = torch.mean(x, 0), torch.var(x, 0)
            scale_init = init_scale / torch.sqrt(v_init + eps)
            self.linear._parameters['weight_g'].data = self.linear._parameters['weight_g'].data * scale_init.view(
                self.linear._parameters['weight_g'].data.size())
            self.linear._parameters['bias'].data = self.linear._parameters['bias'].data - m_init * scale_init
            self.initialized = True + self.initialized
            return scale_init[None, :] * (x - m_init[None, :])


class NormedDense(nn.Module):
    """
    Dense layer with normalization
    """

    def __init__(self, tensor_shape, out_features, weightnorm=True):
        super(NormedDense, self).__init__()
        """
        args:
            tensor_shape (tuple): input tensor shape (B x C x D)
            out_features (int): number of output features
            weight (bool): use weight normalization
        """
        self.register_buffer("initialized", torch.tensor(False))
        self._input_shp = tensor_shape
        self.input_features = int(np.prod(tensor_shape[1:]))
        self._output_shp = (-1, out_features)
        self.linear = nn.Linear(self.input_features, out_features)
        # add batch norm
        if not weightnorm:
            self.weightnorm = False
        else:
            self.weightnorm = True
            self.linear = nn.utils.weight_norm(self.linear, dim=0, name="weight")

    def forward(self, x):
        x = x.view(x.size()[0], -1)
        if not self.initialized:
            self.init_parameters(x)
        x = self.linear(x)
        return x

    @property
    def input_shape(self):
        return self._input_shp

    @property
    def output_shape(self):
        return self._output_shp

    def init_parameters(self, x, init_scale=0.05, eps=1e-8):
        if self.weightnorm:
            # initial values
            self.linear._parameters['weight_v'].data.normal_(mean=0, std=init_scale)
            self.linear._parameters['weight_g'].data.fill_(1.)
            self.linear._parameters['bias'].data.fill_(0.)
            init_scale = .01
            # data dependent init
            x = self.linear(x)
            m_init, v_init = torch.mean(x, 0), torch.var(x, 0)
            scale_init = init_scale / torch.sqrt(v_init + eps)
            self.linear._parameters['weight_g'].data = self.linear._parameters['weight_g'].data * scale_init.view(
                self.linear._parameters['weight_g'].data.size())
            self.linear._parameters['bias'].data = self.linear._parameters['bias'].data - m_init * scale_init
            self.initialized = True + self.initialized
            return scale_init[None, :] * (x - m_init[None, :])


class AsFeatureMap(nn.Module):
    def __init__(self, input_shape, target_shape, weightnorm=True, **kwargs):
        super().__init__()

        self._input_shp = input_shape

        if len(input_shape) < len(target_shape):
            out_features = np.prod(target_shape[1:])
            self.linear = NormedDense(input_shape, out_features, weightnorm=weightnorm)
            self._output_shp = target_shape

        else:
            self.linear = None
            self._output_shp = input_shape

    def forward(self, x):
        if self.linear is None:
            return x

        x = self.linear(x)
        return x.view(self.output_shape)

    @property
    def input_shape(self):
        return self._input_shp

    @property
    def output_shape(self):
        return self._output_shp



import math
from functools import partial
from typing import *

import numpy as np
import torch
from booster import Diagnostic
from torch import Tensor, nn

from .freebits import FreeBits
from ..utils import batch_reduce, log_sum_exp, detach_to_device


class VariationalInference(object):
    def __init__(self, likelihood: any, iw_samples: int = 1, auxiliary: Dict[str, float] = {}, **parameters: Any):
        """
        Initialize VI evaluator
        :param likelihood: likelihood class used to evaluate log p(x | z)
        :param iw_samples: number of importance-weighted samples
        :param auxiliary: dict of auxiliary losses, each key must have a match in the model output, the auxiliary values define the weight of the auxiliary loss in the overall loss
        :param parameters: additional parameters passed to model and evaluator
        """
        super().__init__()

        assert iw_samples > 0
        self._iw_samples = iw_samples
        self._parameters = parameters
        self._auxiliary = auxiliary
        self.likelihood = likelihood

    @staticmethod
    def compute_kls(kls: Union[Tensor, List[Tensor]], freebits: Optional[Union[float, List[float]]], device: str):
        """compute kl and kl to be accounted in the loss"""

        if kls is None or (isinstance(kls, list) and len(kls) == 0):
            _zero = detach_to_device(0., device)
            return _zero, _zero

        # set kls and freebits as lists
        if not isinstance(kls, list):
            kls = [kls]

        if freebits is not None and not isinstance(freebits, list):
            freebits = [freebits for _ in kls]

        # apply freebits to each
        if freebits is not None:
            kls_loss = (FreeBits(fb)(kl) for fb, kl in zip(freebits, kls))
        else:
            kls_loss = kls

        # sum freebit kls
        kls_loss = [batch_reduce(kl)[:, None] for kl in kls_loss]
        kls_loss = batch_reduce(torch.cat(kls_loss, 1))

        # sum kls
        kls = [batch_reduce(kl)[:, None] for kl in kls]
        kls = batch_reduce(torch.cat(kls, 1))

        return kls, kls_loss

    def compute_elbo(self, x, outputs, beta=1.0, freebits=None, **kwargs):

        # Destructuring dict
        x_ = outputs.get('x_')
        kls = outputs.get('kl')

        # compute E_p(x) [ - log p_\theta(x | z) ]
        nll = - batch_reduce(self.likelihood(logits=x_).log_prob(x))

        # compute kl: \sum_i E_q(z_i) [ log q(z_i | h) - log p(z_i | h) ]
        kl, kls_loss = self.compute_kls(kls, freebits, x.device)

        # compute total loss and elbo
        loss = nll + beta * kls_loss
        elbo = -(nll + kl)

        # compute auxiliary losses / kls
        auxiliary = {}
        for k, default_value in self._auxiliary.items():
            # compute value
            value = outputs.get(k, None)
            if value is not None:
                value, _ = self.compute_kls(value, None, x.device)

                # get custom weights from kwargs
                weight = kwargs.get(k, default_value)

                # add to loss
                loss = loss + weight * value

                # store as a tuple
                auxiliary[k] = (weight, value)

        return loss, elbo, kls, kl, nll, auxiliary

    def __call__(self, model: nn.Module, x: Tensor, **kwargs: Any) -> Tuple[Tensor, Dict, Dict]:
        """
        Process inputs using model and compute loss, ELBO and diagnostics.
        :param model: model to evaluate
        :param x: input tensor
        :param kwargs: other args passed both to the model and the evaluator
        :return: (loss, diagnostics)
        """

        # update kwargs
        kwargs.update(self._parameters)

        # importance-weighted placeholders
        iw_elbos = torch.zeros((self._iw_samples, x.size(0)), device=x.device, dtype=torch.float)
        iw_kls = torch.zeros((self._iw_samples, x.size(0)), device=x.device, dtype=torch.float)
        iw_nlls = torch.zeros((self._iw_samples, x.size(0)), device=x.device, dtype=torch.float)

        # Effective Sample size
        # w_i = p(x, z_i) / q(z_i | x)
        ratios = torch.zeros((self._iw_samples, x.size(0)), device=x.device, dtype=torch.float)

        # feed forward pass
        for k in range(self._iw_samples):
            # forward pass
            outputs = model(x, **kwargs)

            # compute VI elbo
            loss, elbo, kls, kl, nll, auxiliary = self.compute_elbo(x, outputs, **kwargs)
            iw_elbos[k, :] = elbo
            iw_kls[k, :] = kl
            iw_nlls[k, :] = nll
            ratios[k, :] = elbo.exp()

        if self._iw_samples > 1:
            elbo = log_sum_exp(iw_elbos, dim=0, sum_op=torch.mean)
            kl = iw_kls.mean(0)
            nll = iw_nlls.mean(0)

        # Compute effective sample size
        N_eff = torch.sum(ratios, 0) ** 2 / torch.sum(ratios ** 2, 0)

        # gather diagnostics
        bits_per_dim = - elbo / math.log(2.) / np.prod(x.size()[1:])
        format = partial(detach_to_device, device=x.device)
        diagnostics = {
            "loss": {"loss": format(loss), "elbo": format(elbo), "kl": format(kl), "nll": format(nll),
                     "bpd": format(bits_per_dim)},
            "info": {"N_eff": format(N_eff), "batch_size": x.size(0)}
        }

        # add auxiliary
        for k, (weight, value) in auxiliary.items():
            diagnostics['loss'][k] = format(value.float().mean())

        # add kls
        diagnostics['kl'] = {f'kl-{i}': v.mean() for i, v in enumerate(kls)}

        # add other params:
        def _check_type(v):
            return isinstance(v, float) or (isinstance(v, Tensor) and v.dim() == 0)

        diagnostics['parameters'] = {k: v for k, v in kwargs.items() if _check_type(v)}

        diagnostics = Diagnostic(diagnostics).to(x.device)

        return loss.mean(), diagnostics, outputs



import math
import os

import matplotlib.image
import numpy as np
import torch
from torchvision.utils import make_grid


def save_img(img, path):
    def _scale(img):
        img *= 255
        return img.astype(np.uint8)

    img = _scale(img)

    matplotlib.image.imsave(path, img)


def summary2logger(logger, summary, global_step, epoch, best=None, stats_key='loss'):
    """write summary to logging"""
    if not stats_key in summary.keys():
        logger.warning('key ' + str(stats_key) + ' not int output dictionary')
    else:
        message = f'\t[{global_step} / {epoch}]   '
        message += ''.join([f'{k} {v:6.2f}   ' for k, v in summary.get(stats_key).items()])
        message += f'({summary["info"]["elapsed-time"]:.2f}s /iter)'
        if best is not None:
            message += f'   (best: {best[0]:6.2f}  [{best[1]} / {best[2]}])'
        logger.info(message)


def save_model(model, eval_summary, global_step, epoch, best_elbo, logdir, key='elbo'):
    elbo = eval_summary['loss'][key]
    prev_elbo, *_ = best_elbo
    if elbo > prev_elbo:
        best_elbo = (elbo, global_step, epoch)
        pth = os.path.join(logdir, "model.pth")
        torch.save(model.state_dict(), pth)

    return best_elbo


def load_model(model, logdir):
    device = next(iter(model.parameters())).device
    model.load_state_dict(torch.load(os.path.join(logdir, "model.pth"), map_location=device))


@torch.no_grad()
def sample_model(model, likelihood, logdir, global_step=0, writer=None, N=100):
    # sample model
    x_ = model.sample_from_prior(N).get('x_')
    x_ = likelihood(logits=x_).sample()

    # make grid
    nrow = math.floor(math.sqrt(N))
    grid = make_grid(x_, nrow=nrow)

    # normalize
    grid -= grid.min()
    grid /= grid.max()

    # log to tensorboard
    if writer is not None:
        writer.add_image('samples', grid, global_step)

    # save the raw image
    img = grid.data.permute(1, 2, 0).cpu().numpy()
    matplotlib.image.imsave(os.path.join(logdir, "samples.png"), img)



from .stochastic import DenseNormal, ConvNormal


def get_deep_vae_mnist():
    """
    Get the binary images Deep VAE configuration.
    :return: enc, z
    """
    filters = 64
    no_layers = 2
    enc = []
    z = []

    enc_z1 = [[filters, 5, 1]] * no_layers
    enc_z1 += [[filters, 5, 2]]
    z1 = {'N': 48, 'block': DenseNormal}
    enc += [enc_z1]
    z += [z1]

    enc_z2 = [[filters, 3, 1]] * no_layers
    enc_z2 += [[filters, 3, 1]]
    z2 = {'N': 40, 'block': DenseNormal}
    enc += [enc_z2]
    z += [z2]

    enc_z3 = [[filters, 3, 1]] * no_layers
    enc_z3 += [[filters, 3, 1]]
    z3 = {'N': 32, 'block': DenseNormal}
    enc += [enc_z3]
    z += [z3]

    enc_z4 = [[filters, 3, 1]] * no_layers
    enc_z4 += [[filters, 3, 1]]
    z4 = {'N': 24, 'block': DenseNormal}
    enc += [enc_z4]
    z += [z4]

    enc_z5 = [[filters, 3, 1]] * no_layers
    enc_z5 += [[filters, 3, 1]]
    z5 = {'N': 16, 'block': DenseNormal}
    enc += [enc_z5]
    z += [z5]

    enc_z6 = [[filters, 3, 1]] * no_layers
    enc_z6 += [[filters, 3, 2]]
    z6 = {'N': 8, 'block': DenseNormal}
    enc += [enc_z6]
    z += [z6]

    return enc, z


def get_deep_vae_cifar():
    filters = 96
    no_layers = 2
    enc = []
    z = []

    enc_z1 = [[filters, 5, 1]] * no_layers
    enc_z1 += [[filters, 5, 2]]
    z_1 = {'N': 38, 'kernel': 16, 'block': ConvNormal}
    enc += [enc_z1]
    z += [z_1]

    enc_z2 = [[filters, 3, 1]] * no_layers
    enc_z2 += [[filters, 3, 1]]
    z_2 = {'N': 36, 'kernel': 16, 'block': ConvNormal}
    enc += [enc_z2]
    z += [z_2]

    enc_z3 = [[filters, 3, 1]] * no_layers
    enc_z3 += [[filters, 3, 1]]
    z_3 = {'N': 34, 'kernel': 16, 'block': ConvNormal}
    enc += [enc_z3]
    z += [z_3]

    enc_z4 = [[filters, 3, 1]] * no_layers
    enc_z4 += [[filters, 3, 1]]
    z_4 = {'N': 32, 'kernel': 16, 'block': ConvNormal}
    enc += [enc_z4]
    z += [z_4]

    enc_z5 = [[filters, 3, 1]] * no_layers
    enc_z5 += [[filters, 3, 1]]
    z_5 = {'N': 30, 'kernel': 16, 'block': ConvNormal}
    enc += [enc_z5]
    z += [z_5]

    enc_z6 = [[filters, 3, 1]] * no_layers
    enc_z6 += [[filters, 3, 1]]
    z_6 = {'N': 28, 'kernel': 16, 'block': ConvNormal}
    enc += [enc_z6]
    z += [z_6]

    enc_z7 = [[filters, 3, 1]] * no_layers
    enc_z7 += [[filters, 3, 1]]
    z_7 = {'N': 26, 'kernel': 16, 'block': ConvNormal}
    enc += [enc_z7]
    z += [z_7]

    enc_z8 = [[filters, 3, 1]] * no_layers
    enc_z8 += [[filters, 3, 1]]
    z_8 = {'N': 24, 'kernel': 16, 'block': ConvNormal}
    enc += [enc_z8]
    z += [z_8]

    enc_z9 = [[filters, 3, 1]] * no_layers
    enc_z9 += [[filters, 3, 1]]
    z_9 = {'N': 22, 'kernel': 16, 'block': ConvNormal}
    enc += [enc_z9]
    z += [z_9]

    enc_z10 = [[filters, 3, 1]] * no_layers
    enc_z10 += [[filters, 3, 1]]
    z_10 = {'N': 20, 'kernel': 16, 'block': ConvNormal}
    enc += [enc_z10]
    z += [z_10]

    enc_z11 = [[filters, 3, 1]] * no_layers
    enc_z11 += [[filters, 3, 2]]
    z_11 = {'N': 18, 'kernel': 8, 'block': ConvNormal}
    enc += [enc_z11]
    z += [z_11]

    enc_z12 = [[filters, 3, 1]] * no_layers
    enc_z12 += [[filters, 3, 1]]
    z_12 = {'N': 16, 'kernel': 8, 'block': ConvNormal}
    enc += [enc_z12]
    z += [z_12]

    enc_z13 = [[filters, 3, 1]] * no_layers
    enc_z13 += [[filters, 3, 1]]
    z_13 = {'N': 14, 'kernel': 8, 'block': ConvNormal}
    enc += [enc_z13]
    z += [z_13]

    enc_z14 = [[filters, 3, 1]] * no_layers
    enc_z14 += [[filters, 3, 1]]
    z_14 = {'N': 12, 'kernel': 8, 'block': ConvNormal}
    enc += [enc_z14]
    z += [z_14]

    enc_z15 = [[filters, 3, 1]] * no_layers
    enc_z15 += [[filters, 3, 2]]
    z_15 = {'N': 10, 'kernel': 4, 'block': ConvNormal}
    enc += [enc_z15]
    z += [z_15]

    return enc, z



from .architectures import get_deep_vae_mnist, get_deep_vae_cifar
from .deepvae import DeepVae, VAE, LVAE, BIVA
from .stage import BaseStage, VaeStage, LvaeStage, BivaStage
from .stochastic import StochasticLayer, DenseNormal, ConvNormal



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='biva-pytorch',
    version='0.1.4',
    author="Valentin Lievin",
    author_email="valentin.lievin@gmail.com",
    description="Official PyTorch BIVA implementation (BIVA: A Very Deep Hierarchy of Latent Variables for Generative Modeling)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vlievin/biva-pytorch",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'torch',
        'tqdm',
        'numpy',
        'matplotlib',
        'tensorboard',
        'booster-pytorch'
    ],
)



from copy import copy
from typing import *

import torch
from torch import nn, Tensor

from .stochastic import StochasticLayer
from .utils import DataCollector
from ..layers import GatedResNet, AsFeatureMap
from ..utils import shp_cat


def StochasticBlock(data: Dict, *args, **kwargs):
    """Construct the stochastic block given by the Block argument"""
    Block = data.get('block')
    block = Block(data, *args, **kwargs)
    assert isinstance(block, StochasticLayer)
    return block


class DeterministicBlocks(nn.Module):

    def __init__(self,
                 tensor_shp: Tuple[int],
                 convolutions: List[Tuple[int]],
                 in_residual: bool = True,
                 transposed: bool = False,
                 Block: Any = GatedResNet,
                 aux_shape: Optional[List[Tuple[int]]] = None,
                 **kwargs):
        """
        Defines a of sequence of deterministic blocks (resnets).
        You can extend this class by passing other Block classes as an argument.

        auxiliary connections: if the number of auxiliary inputs is smaller than the number of layers,
        the auxiliary inputs are repeated to match the number of layers.

        :param tensor_shp: input tensor shape as a tuple of integers (B, H, *D)
        :param convolutions: describes the sequence of blocks, each of them defined by a tuple  (filters, kernel_size, stride)
        :param aux_shape: auxiliary input tensor shape as a tuple of integers (B, H, *D)
        :param transposed: use transposed convolutions
        :param residual: use residual connections
        :param Block: Block object constructor (GatedResNet, ResMLP)
        """
        super().__init__()
        self.input_shape = tensor_shp
        self._use_skips = True
        layers = []

        if aux_shape is None:
            self._use_skips = False
            aux_shape = []

        for j, dim in enumerate(convolutions):
            residual = True if j > 0 else in_residual
            aux = aux_shape.pop() if self._use_skips else None
            block = Block(tensor_shp, dim, aux_shape=aux, transposed=transposed, residual=residual,
                          **kwargs)
            tensor_shp = block.output_shape
            aux_shape = [tensor_shp] + aux_shape
            layers += [block]

        self.layers = nn.ModuleList(layers)
        self.output_shape = tensor_shp
        self.hidden_shapes = aux_shape

    def __len__(self):
        return len(self.layers)

    def forward(self, x: Tensor, aux: Optional[List[Tensor]] = None, **kwargs) -> Tuple[Tensor, List[Tensor]]:
        """
        :param x: input tensor
        :param aux: list of auxiliary inputs
        :return: output tensor, activations
        """
        if aux is None:
            aux = []

        for layer in self.layers:
            a = aux.pop() if self._use_skips else None
            x = layer(x, a, **kwargs)
            aux = [x] + aux

        return x, aux


class BaseStage(nn.Module):
    def __init__(self,
                 input_shape: Dict[str, Tuple[int]],
                 convolutions: List[Tuple[int]],
                 stochastic: Tuple,
                 top: bool = False,
                 bottom: bool = False,
                 q_dropout: float = 0,
                 p_dropout: float = 0,
                 Block: Any = GatedResNet,
                 no_skip: bool = False,
                 **kwargs):
        """
        Define a stage of a hierarchical model.
        In a VAE setting, a stage defines:
        * the latent variable z_i
        * the encoder q(z_i | h_{q<i})
        * the decoder p(z_{i-1} | z_i)
        """
        super().__init__()
        self._input_shape = input_shape
        self._convolutions = convolutions
        self._stochastic = stochastic
        self._top = top
        self._bottom = bottom
        self._no_skip = no_skip

    @property
    def input_shape(self) -> Dict[str, Tuple[int]]:
        """size of the input tensors for the inference path"""
        return self._input_shape

    @property
    def q_output_shape(self) -> Dict[str, Tuple[int]]:
        """size of the output tensors for the inference path"""
        raise NotImplementedError

    @property
    def forward_shape(self) -> Tuple[int]:
        """size of the output tensor for the generative path"""
        raise NotImplementedError

    def infer(self, data: Dict[str, Tensor], **kwargs) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Perform a forward pass through the inference layers and sample the posterior.

        :param data: input data
        :param kwargs: additional parameters passed to the stochastic layer
        :return: (output data, variational data)
        """
        raise NotImplementedError

    def forward(self, data: dict, posterior: Optional[dict], **kwargs) -> Tuple[
        Dict, Dict[str, List]]:
        """
        Perform a forward pass through the generative model and compute KL if posterior data is available

        :param data: data from the above stage forward pass
        :param posterior: dictionary representing the posterior from same stage inference pass
        :return: (dict('d' : d, 'aux : [aux]), dict('kl': [kl], **auxiliary) )
        """
        raise NotImplementedError


class VaeStage(BaseStage):
    def __init__(self,
                 input_shape: Dict[str, Tuple[int]],
                 convolutions: List[Tuple[int]],
                 stochastic: Tuple,
                 top: bool = False,
                 bottom: bool = False,
                 q_dropout: float = 0,
                 p_dropout: float = 0,
                 Block: Any = GatedResNet,
                 no_skip: bool = False,
                 **kwargs):
        """
        VAE: https://arxiv.org/abs/1312.6114

        Define a Variational Autoencoder stage containing:
        - a sequence of convolutional blocks for the inference model
        - a sequence of convolutional blocks for the generative model
        - a stochastic layer

        :param input_shape: dictionary describing the input tensors of shapes (B, H, *D)
        :param convolution: list of tuple describing a convolutional block (filters, kernel_size, stride)
        :param stochastic: integer or tuple describing the stochastic layer: units or (units, kernel_size, discrete, K)
        :param top: is top layer
        :param bottom: is bottom layer
        :param q_dropout: inference dropout value
        :param p_dropout: generative dropout value
        :param Block: Block constructor
        :param no_skip: do not use skip connections
        :param kwargs: others arguments passed to the block constructors (both convolutions and stochastic)
        """
        super().__init__(input_shape, convolutions, stochastic, top=top, bottom=bottom, q_dropout=q_dropout,
                         p_dropout=p_dropout, Block=Block, no_skip=no_skip)

        tensor_shp = input_shape.get('x')
        aux_shape = input_shape.get('aux', None)

        # mute skip connections
        if no_skip:
            aux_shape = None

        # define inference convolutional blocks
        in_residual = not bottom
        q_skips = [aux_shape for _ in convolutions] if aux_shape is not None else None
        self.q_convs = DeterministicBlocks(tensor_shp, convolutions, aux_shape=q_skips, transposed=False,
                                           in_residual=in_residual, Block=Block, dropout=q_dropout, **kwargs)

        # shape of the deterministic output
        tensor_shp = self.q_convs.output_shape

        # define the stochastic layer
        self.stochastic = StochasticBlock(stochastic, tensor_shp, top=top, **kwargs)
        self.q_proj = AsFeatureMap(self.stochastic.output_shape, self.stochastic.input_shape)

        self._q_output_shape = {'x': self.q_proj.output_shape, 'aux': tensor_shp}

        ### GENERATIVE MODEL

        # project z sample
        self.p_proj = AsFeatureMap(self.stochastic.output_shape, self.stochastic.input_shape)

        # define the generative convolutional blocks with the skip connections
        # here we assume the skip connections to be of the same shape as `tensor_shp` : this does not work with
        # with every configuration of the generative model. Making the arhitecture more general requires to have
        # a top-down __init__() method such as to take the shapes of the above generative block skip connections as input.
        p_skips = None if (top or no_skip) else [tensor_shp] * len(convolutions)
        self.p_convs = DeterministicBlocks(self.p_proj.output_shape, convolutions[::-1],
                                           aux_shape=p_skips, transposed=True,
                                           in_residual=False, Block=Block, dropout=p_dropout, **kwargs)

        self._p_output_shape = {'d': self.p_convs.output_shape, 'aux': self.p_convs.hidden_shapes}

    @property
    def q_output_shape(self) -> Dict[str, Tuple[int]]:
        """size of the output tensors for the inference path"""
        return self._q_output_shape

    @property
    def p_output_shape(self) -> Tuple[int]:
        """size of the output tensor for the generative path"""
        return self._p_output_shape

    def infer(self, data: Dict[str, Tensor], **kwargs) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Perform a forward pass through the inference layers and sample the posterior.

        :param data: input data
        :param kwargs: additional parameters passed to the stochastic layer
        :return: (output data, variational data)
        """
        x = data.get('x')
        aux = data.get('aux', None)
        if self._no_skip:
            aux = None

        aux = [aux for _ in range(len(self.q_convs))] if aux is not None else None
        x, _ = self.q_convs(x, aux)

        z, q_data = self.stochastic(x, inference=True, **kwargs)
        z = self.q_proj(z)

        return {'x': z, 'aux': x}, q_data

    def forward(self, data: dict, posterior: Optional[dict], **kwargs) -> Tuple[
        Dict, Dict[str, List]]:
        """
        Perform a forward pass through the generative model and compute KL if posterior data is available

        :param data: data from the above stage forward pass
        :param posterior: dictionary representing the posterior from same stage inference pass
        :return: (dict('d' : d, 'aux : [aux]), dict('kl': [kl], **auxiliary) )
        """
        d = data.get('d', None)
        aux = data.get('aux', None)
        if self._no_skip:
            aux = None

        # sample p(z | d)
        z_p, p_data = self.stochastic(d, inference=False, sample=posterior is None, **kwargs)

        # compute KL(q | p)
        if posterior is not None:
            loss_data = self.stochastic.loss(posterior, p_data, **kwargs)
            z = posterior.get('z')
        else:
            loss_data = {}
            z = z_p

        # project z
        z = self.p_proj(z)

        # pass through convolutions
        d, skips = self.p_convs(z, aux=aux)

        output_data = {'d': d, 'aux': skips}
        return output_data, loss_data


class LvaeStage(VaeStage):
    def __init__(self,
                 input_shape: Dict[str, Tuple[int]],
                 convolutions: List[Tuple[int]],
                 stochastic: Tuple,
                 top: bool = False,
                 bottom: bool = False,
                 q_dropout: float = 0,
                 p_dropout: float = 0,
                 Block: Any = GatedResNet,
                 **kwargs):
        """
        LVAE: https://arxiv.org/abs/1602.02282

        Define a Ladder Variational Autoencoder stage containing:
        - a sequence of convolutional blocks for the inference model
        - a sequence of convolutional blocks for the generative model
        - a stochastic layer

        :param input_shape: dictionary describing the input tensors of shapes (B, H, *D)
        :param convolution: list of tuple describing a convolutional block (filters, kernel_size, stride)
        :param stochastic: integer or tuple describing the stochastic layer: units or (units, kernel_size, discrete, K)
        :param top: is top layer
        :param bottom: is bottom layer
        :param q_dropout: inference dropout value
        :param p_dropout: generative dropout value
        :param kwargs: others arguments passed to the block constructors (both convolutions and stochastic)
        """
        super().__init__(input_shape, convolutions, stochastic, top=top, bottom=bottom, p_dropout=p_dropout,
                         q_dropout=q_dropout, Block=Block, **kwargs)
        self.q_proj = None
        # get the tensor shape of the output of the deterministic path
        top_shape = self._q_output_shape.get('aux')
        # modify the output of the inference path to be only deterministic
        self._q_output_shape['x'] = top_shape

        topdown = top_shape if not top else None
        conv = convolutions[-1]
        if isinstance(conv, list):
            conv = [conv[0], conv[1], 1, conv[-1]]
        self.merge = Block(top_shape, conv, aux_shape=topdown, transposed=False, in_residual=True, dropout=p_dropout,
                           **kwargs)

    def infer(self, data: Dict[str, Tensor], **kwargs) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Perform a forward pass through the inference layers and sample the posterior.

        :param data: input data
        :param kwargs: additional parameters passed to the stochastic layer
        :return: (output data, variational data)
        """
        x = data.get('x')
        aux = data.get('aux', None)
        if self._no_skip:
            aux = None

        aux = [aux for _ in range(len(self.q_convs))] if aux is not None else None
        x, _ = self.q_convs(x, aux)

        return {'x': x, 'aux': x}, {'h': x}

    def forward(self, data: dict, posterior: Optional[dict], debugging: bool = False, **kwargs) -> Tuple[
        Dict, Dict[str, List[Tensor]]]:
        """
        Perform a forward pass through the generative model and compute KL if posterior data is available

        :param data: data from the above stage forward pass
        :param posterior: dictionary representing the posterior
        :return: (dict('d' : d, 'aux : [aux]), dict('kl': [kl], **auxiliary) )
        """
        d = data.get('d', None)

        # sample p(z | d)
        z_p, p_data = self.stochastic(d, inference=False, sample=posterior is None, **kwargs)

        # sample q(z | h) and compute KL(q | p)
        if posterior is not None:
            # compute the top-down logits of q(z_i | x, z_{>i})
            h = posterior.get('h')
            h = self.merge(h, aux=d)

            # z ~ q(z | h_bu, d_td)
            z_q, q_data = self.stochastic(h, inference=True, **kwargs)

            loss_data = self.stochastic.loss(q_data, p_data, **kwargs)
            z = z_q
        else:
            loss_data = {}
            z = z_p

        # project z
        z = self.p_proj(z)

        # pass through convolutions
        aux = data.get('aux', None)
        if self._no_skip:
            aux = None

        d, skips = self.p_convs(z, aux)

        output_data = {'d': d, 'aux': skips}
        return output_data, loss_data


class BivaIntermediateStage(BaseStage):
    def __init__(self,
                 input_shape: Dict[str, Tuple[int]],
                 convolutions: List[Tuple[int]],
                 stochastic: Union[Dict, Tuple[Dict]],
                 top: bool = False,
                 bottom: bool = False,
                 q_dropout: float = 0,
                 p_dropout: float = 0,
                 no_skip: bool = False,
                 conditional_bu: bool = False,
                 Block: Any = GatedResNet,
                 merge_kernel: int = 3,
                 **kwargs):
        """
        BIVA: https://arxiv.org/abs/1902.02102

        Define a Bidirectional Variational Autoencoder stage containing:
        - a sequence of convolutional blocks for the bottom-up inference model (BU)
        - a sequence of convolutional blocks for the top-down inference model (TD)
        - a sequence of convolutional blocks for the generative model
        - two stochastic layers (BU and TD)

        :param input_shape: dictionary describing the input tensor shape (B, H, *D)
        :param convolution: list of tuple describing a convolutional block (filters, kernel_size, stride)
        :param stochastic: dictionary describing the stochastic layer: units or (units, kernel_size, discrete, K)
        :param bottom: is bottom layer
        :param top: is top layer
        :param q_dropout: inference dropout value
        :param p_dropout: generative dropout value
        :param no_skip: do not use skip connections
        :param conditional_bu: condition BU prior on p(z_TD)
        :param aux_shape: auxiliary input tensor shape as a tuple of integers (B, H, *D)
        :param kwargs: others arguments passed to the block constructors (both convolutions and stochastic)
        """
        super().__init__(input_shape, convolutions, stochastic, top=top, bottom=bottom, q_dropout=q_dropout,
                         p_dropout=p_dropout, Block=Block, no_skip=no_skip)

        self._conditional_bu = conditional_bu
        self._merge_kernel = merge_kernel

        if 'x' in input_shape.keys():
            bu_shp = td_shp = input_shape.get('x')
            aux_shape = None
        else:
            bu_shp = input_shape.get('x_bu')
            td_shp = input_shape.get('x_td')
            aux_shape = input_shape.get('aux')

        if isinstance(stochastic.get('block'), tuple):
            bu_block, td_block = stochastic.get('block')
            bu_stochastic = copy(stochastic)
            td_stochastic = copy(stochastic)
            bu_stochastic['block'] = bu_block
            td_stochastic['block'] = td_block
        else:
            bu_stochastic = td_stochastic = stochastic

        # mute skip connections
        if no_skip:
            aux_shape = None

        # define inference convolutional blocks
        in_residual = not bottom
        q_bu_aux = [aux_shape for _ in convolutions] if aux_shape is not None else None
        self.q_bu_convs = DeterministicBlocks(bu_shp, convolutions, aux_shape=q_bu_aux, transposed=False,
                                              in_residual=in_residual, dropout=q_dropout, Block=Block, **kwargs)

        q_td_aux = [self.q_bu_convs.output_shape for _ in convolutions]
        self.q_td_convs = DeterministicBlocks(td_shp, convolutions, aux_shape=q_td_aux, transposed=False,
                                              in_residual=in_residual, dropout=q_dropout, Block=Block, **kwargs)

        # shape of the output of the inference path and input tensor from the generative path
        top_tensor_shp = self.q_td_convs.output_shape
        aux_shape = shp_cat([top_tensor_shp, top_tensor_shp], 1)

        # define the BU stochastic layer
        bu_top = False if conditional_bu else top
        self.bu_stochastic = StochasticBlock(bu_stochastic, top_tensor_shp, top=bu_top, **kwargs)
        self.bu_proj = AsFeatureMap(self.bu_stochastic.output_shape, self.bu_stochastic.input_shape, **kwargs)

        # define the TD stochastic layer
        self.td_stochastic = StochasticBlock(td_stochastic, top_tensor_shp, top=top, **kwargs)

        self._q_output_shape = {'x_bu': self.bu_proj.output_shape,
                                'x_td': top_tensor_shp,
                                'aux': aux_shape}

        ### GENERATIVE MODEL

        # TD merge layer
        h_shape = self._q_output_shape.get('x_td', None) if not self._top else None
        conv = self._convolutions[::-1][-1]
        if isinstance(conv, list) or isinstance(conv, tuple):
            conv = [conv[0], merge_kernel, 1,
                    conv[-1]]  # in the original implementation, this depends on the parameters of the above layers
        self.merge = Block(h_shape, conv, aux_shape=h_shape, transposed=False, in_residual=True,
                           dropout=p_dropout,
                           **kwargs)

        # alternative: define the condition p(z_bu | z_td, ...)
        if conditional_bu:
            self.bu_condition = Block(self.bu_stochastic.output_shape, conv, aux_shape=h_shape, transposed=False,
                                      in_residual=False,
                                      dropout=p_dropout, **kwargs)
        else:
            self.bu_condition = None

        # merge latent variables
        z_shp = shp_cat([self.bu_stochastic.output_shape, self.td_stochastic.output_shape], 1)
        self.z_proj = AsFeatureMap(z_shp, self.bu_stochastic.input_shape)

        # define the generative convolutional blocks with the skip connections
        # here we assume the skip connections to be of the same shape as `top_tensor_shape` : this does not work with
        # with every configuration of the generative model. Making the arhitecture more general requires to have
        # a top-down __init__() method such as to take the shapes of the above generative block skip connections as input.
        p_skips = None if (top or no_skip) else [top_tensor_shp] * len(convolutions)
        self.p_convs = DeterministicBlocks(self.z_proj.output_shape, self._convolutions[::-1],
                                           aux_shape=p_skips, transposed=True,
                                           in_residual=False, Block=Block, dropout=p_dropout, **kwargs)

        self._p_output_shape = {'d': self.p_convs.output_shape, 'aux': self.p_convs.hidden_shapes}

    @property
    def q_output_shape(self) -> Dict[str, Tuple[int]]:
        """size of the output tensors for the inference path"""
        return self._q_output_shape

    @property
    def p_output_shape(self) -> Tuple[int]:
        """size of the output tensor for the generative path"""
        return self._p_output_shape

    def infer(self, data: Dict[str, Tensor], **kwargs) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Perform a forward pass through the inference layers and sample the posterior.

        :param data: input data
        :param kwargs: additional parameters passed to the stochastic layer
        :return: (output data, variational data)
        """
        if 'x' in data.keys():
            x = data.get('x')
            x_bu, x_td = x, x
        else:
            x_bu = data.get('x_bu')
            x_td = data.get('x_td')

        aux = data.get('aux', None)
        if self._no_skip:
            aux = None

        # BU path
        bu_aux = [aux for _ in range(len(self.q_bu_convs))] if aux is not None else None
        x_bu, _ = self.q_bu_convs(x_bu, aux=bu_aux)
        # z_bu ~ q(x)
        z_bu, bu_q_data = self.bu_stochastic(x_bu, inference=True, **kwargs)
        z_bu_proj = self.bu_proj(z_bu)

        # TD path
        td_aux = [x_bu for _ in range(len(self.q_td_convs))]
        x_td, _ = self.q_td_convs(x_td, aux=td_aux)
        td_q_data = {'z': z_bu, 'h': x_td}  # note h = d_q(x)

        # skip connection
        aux = torch.cat([x_bu, x_td], 1)

        return {'x_bu': z_bu_proj, 'x_td': x_td, 'aux': aux}, {'z_bu': z_bu, 'bu': bu_q_data, 'td': td_q_data}

    def forward(self, data: Dict, posterior: Optional[dict], debugging: bool = False, **kwargs) -> Tuple[
        Dict, Dict[str, List[Tensor]]]:
        """
        Perform a forward pass through the generative model and compute KL if posterior data is available

        :param d: previous hidden state
        :param posterior: dictionary representing the posterior
        :return: (hidden state, dict('kl': [kl], **auxiliary))
        """

        d = data.get('d', None)

        if posterior is not None:
            # sample posterior and compute KL using prior
            bu_q_data = posterior.get('bu')
            td_q_data = posterior.get('td')
            z_bu_q = posterior.get('z_bu')

            # top-down: compute the posterior using the bottom-up hidden state and top-down hidden state
            # p(z_td | d_top)
            _, td_p_data = self.td_stochastic(d, inference=False, sample=False, **kwargs)

            # merge d_top with h = d_q(x)
            h = td_q_data.get('h')
            h = self.merge(h, aux=d)

            # z_td ~ q(z_td | h_bu_td)
            z_td_q, td_q_data = self.td_stochastic(h, inference=True, sample=True, **kwargs)

            # compute log q_bu(z_i | x) - log p_bu(z_i) (+ additional data)
            td_loss_data = self.td_stochastic.loss(td_q_data, td_p_data, **kwargs)

            # conditional BU prior
            if self.bu_condition is not None:
                d_ = self.bu_condition(z_td_q, aux=d)
            else:
                d_ = d

            # bottom-up: retrieve data from the inference path
            # z_bu ~ p(d_top)
            _, bu_p_data = self.bu_stochastic(d_, inference=False, sample=False, **kwargs)

            # compute log q_td(z_i | x, z_{>i}) - log p_td(z_i) (+ additional data)
            bu_loss_data = self.bu_stochastic.loss(bu_q_data, bu_p_data, **kwargs)

            # merge samples
            z = torch.cat([z_td_q, z_bu_q], 1)

        else:
            # sample priors
            # top-down
            z_td_p, td_p_data = self.td_stochastic(d, inference=False, sample=True, **kwargs)  # prior

            # conditional BU prior
            if self.bu_condition is not None:
                d_ = self.bu_condition(z_td_p, aux=d)
            else:
                d_ = d

            # bottom-up
            z_bu_p, bu_p_data = self.bu_stochastic(d_, inference=False, sample=True, **kwargs)  # prior

            bu_loss_data, td_loss_data = {}, {}

            # merge samples
            z = torch.cat([z_td_p, z_bu_p], 1)

        # projection
        z = self.z_proj(z)

        # pass through convolutions
        aux = data.get('aux', None)
        if self._no_skip:
            aux = None

        d, skips = self.p_convs(z, aux=aux)

        # gather data
        loss_data = DataCollector()
        loss_data.extend(td_loss_data)
        loss_data.extend(bu_loss_data)

        output_data = {'d': d, 'aux': skips}
        return output_data, loss_data


class BivaTopStage_simpler(VaeStage):
    """
    This is the BivaTopStage without the additional BU-TD merge layer.
    """

    def __init__(self, input_shape: Dict[str, Tuple[int]], *args, **kwargs):
        bu_shp = input_shape.get('x_bu')
        td_shp = input_shape.get('x_td')

        tensor_shp = shp_cat([bu_shp, td_shp], 1)
        concat_shape = {'x': tensor_shp}

        super().__init__(concat_shape, *args, **kwargs)

    def infer(self, data: Dict[str, Tensor], **kwargs) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        x_bu = data.pop('x_bu')
        x_td = data.pop('x_td')
        data['x'] = torch.cat([x_bu, x_td], 1)

        return super().infer(data, **kwargs)


class BivaTopStage(BaseStage):
    def __init__(self, input_shape: Dict[str, Tuple[int]],
                 convolutions: List[Tuple[int]],
                 stochastic: Union[Dict, Tuple[Dict]],
                 top: bool = False,
                 bottom: bool = False,
                 q_dropout: float = 0,
                 p_dropout: float = 0,
                 no_skip: bool = False,
                 Block: Any = GatedResNet,
                 **kwargs):
        """
        BIVA: https://arxiv.org/abs/1902.02102

        Define a Bidirectional Variational Autoencoder top stage containing:
        - a sequence of convolutional blocks for the bottom-up inference model (BU)
        - a sequence of convolutional blocks for the top-down inference model (TD)
        - a convolutional block to merge BU and TD
        - a sequence of convolutional blocks for the generative model
        - a stochastic layer (z_L)

        :param input_shape: dictionary describing the input tensor shape (B, H, *D)
        :param convolution: list of tuple describing a convolutional block (filters, kernel_size, stride)
        :param stochastic: dictionary describing the stochastic layer: units or (units, kernel_size, discrete, K)
        :param bottom: is bottom layer
        :param top: is top layer
        :param q_dropout: inference dropout value
        :param p_dropout: generative dropout value
        :param no_skip: do not use skip connections
        :param aux_shape: auxiliary input tensor shape as a tuple of integers (B, H, *D)
        :param kwargs: others arguments passed to the block constructors (both convolutions and stochastic)
        """
        super().__init__(input_shape, convolutions, stochastic, top=top, bottom=bottom, q_dropout=q_dropout,
                         p_dropout=p_dropout, Block=Block, no_skip=no_skip)
        top = True

        if 'x' in input_shape.keys():
            bu_shp = td_shp = input_shape.get('x')
            aux_shape = None
        else:
            bu_shp = input_shape.get('x_bu')
            td_shp = input_shape.get('x_td')
            aux_shape = input_shape.get('aux')

        # mute skip connections
        if no_skip:
            aux_shape = None

        # define inference BU and TD paths
        in_residual = not bottom
        q_bu_aux = [aux_shape for _ in convolutions] if aux_shape is not None else None
        self.q_bu_convs = DeterministicBlocks(bu_shp, convolutions, aux_shape=q_bu_aux, transposed=False,
                                              in_residual=in_residual, dropout=q_dropout, Block=Block, **kwargs)

        q_td_aux = [self.q_bu_convs.output_shape for _ in convolutions]
        self.q_td_convs = DeterministicBlocks(td_shp, convolutions, aux_shape=q_td_aux, transposed=False,
                                              in_residual=in_residual, dropout=q_dropout, Block=Block, **kwargs)

        # merge BU and TD paths
        conv = convolutions[-1]
        self.q_top = Block(shp_cat([self.q_bu_convs.output_shape, self.q_td_convs.output_shape], 1),
                           [conv[0], conv[1], 1, conv[-1]], dropout=q_dropout,
                           residual=True, **kwargs)
        top_tensor_shp = self.q_top.output_shape

        # stochastic layer
        self.stochastic = StochasticBlock(stochastic, top_tensor_shp, top=top, **kwargs)

        self._q_output_shape = {}  # no output shape (top layer)

        ### GENERATIVE MODEL

        # map sample back to a feature map
        self.z_proj = AsFeatureMap(self.stochastic.output_shape, self.stochastic.input_shape)

        # define the generative convolutional blocks with the skip connections
        p_skips = None
        self.p_convs = DeterministicBlocks(self.z_proj.output_shape, self._convolutions[::-1],
                                           aux_shape=p_skips, transposed=True,
                                           in_residual=False, Block=Block, dropout=p_dropout, **kwargs)

        self._p_output_shape = {'d': self.p_convs.output_shape, 'aux': self.p_convs.hidden_shapes}

    def infer(self, data: Dict[str, Tensor], **kwargs) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Perform a forward pass through the inference layers and sample the posterior.

        :param data: input data
        :param kwargs: additional parameters passed to the stochastic layer
        :return: (output data, variational data)
        """

        if 'x' in data.keys():
            x = data.get('x')
            x_bu, x_td = x, x
        else:
            x_bu = data.get('x_bu')
            x_td = data.get('x_td')

        aux = data.get('aux', None)
        if self._no_skip:
            aux = None

        # BU path
        bu_aux = [aux for _ in range(len(self.q_bu_convs))] if aux is not None else None
        x_bu, _ = self.q_bu_convs(x_bu, aux=bu_aux)

        # TD path
        td_aux = [x_bu for _ in range(len(self.q_td_convs))]
        x_td, _ = self.q_td_convs(x_td, aux=td_aux)

        # merge BU and TD
        x = torch.cat([x_bu, x_td], 1)
        x = self.q_top(x)

        # sample top layer
        z, q_data = self.stochastic(x, inference=True, **kwargs)

        return {}, q_data

    def forward(self, data: Dict, posterior: Optional[dict], **kwargs) -> Tuple[
        Dict, Dict[str, List]]:
        """
        Perform a forward pass through the generative model and compute KL if posterior data is available

        :param data: data from the above stage forward pass
        :param posterior: dictionary representing the posterior
        :return: (hidden state, dict('kl': [kl], **auxiliary) )
        """
        d = data.get('d', None)

        if posterior is not None:
            # get p(z | d)
            _, p_data = self.stochastic(d, inference=False, sample=False, **kwargs)

            # compute KL(q | p)
            loss_data = self.stochastic.loss(posterior, p_data, **kwargs)
            z = posterior.get('z')
        else:
            loss_data = {}
            z, p_data = self.stochastic(d, inference=False, sample=True, **kwargs)

        # project z
        z = self.z_proj(z)

        # pass through convolutions
        aux = data.get('aux', None)
        if self._no_skip:
            aux = None

        d, skips = self.p_convs(z, aux=aux)

        output_data = {'d': d, 'aux': skips}
        return output_data, loss_data

    @property
    def q_output_shape(self) -> Dict[str, Tuple[int]]:
        """size of the output tensors for the inference path"""
        return self._q_output_shape

    @property
    def p_output_shape(self) -> Tuple[int]:
        """size of the output tensor for the generative path"""
        return self._p_output_shape


def BivaStage(input_shape: Dict[str, Tuple[int]],
              convolutions: List[Tuple[int]],
              stochastic: Union[Dict, Tuple[Dict]],
              top: bool = False,
              **kwargs):
    """
    BIVA: https://arxiv.org/abs/1902.02102

    Define a Bidirectional Variational Autoencoder stage containing:
    - a sequence of convolutional blocks for the bottom-up inference model (BU)
    - a sequence of convolutional blocks for the top-down inference model (TD)
    - a sequence of convolutional blocks for the generative model
    - two stochastic layers (BU and TD)

    This is not an op-for-op implementation of the original Tensorflow version.

    :param input_shape: dictionary describing the input tensor shape (B, H, *D)
    :param convolution: list of tuple describing a convolutional block (filters, kernel_size, stride)
    :param stochastic: dictionary describing the stochastic layer: units or (units, kernel_size, discrete, K)
    :param top: is top layer
    :param bottom: is bottom layer
    :param q_dropout: inference dropout value
    :param p_dropout: generative dropout value
    :param conditional_bu: condition BU prior on p(z_TD)
    :param aux_shape: auxiliary input tensor shape as a tuple of integers (B, H, *D)
    :param kwargs: others arguments passed to the block constructors (both convolutions and stochastic)
    """

    if top:
        return BivaTopStage(input_shape, convolutions, stochastic, top=top, **kwargs)
    else:
        return BivaIntermediateStage(input_shape, convolutions, stochastic, top=top, **kwargs)
